# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

__all__ = ["ClassicalRegister", "QuantumRegister"]

class QuantumRegister:
    """A class to represent a collection of qubits.

    Args:
        start: The starting index of the quantum register.
        size: The number of qubits in the quantum register.
        name: The name of the quantum register. A name will be generated if not provided.
    """

    def __init__(self, start: int, size: int, name: str = "") -> None: ...
    @property
    def start(self) -> int:
        """The index of the first qubit in the quantum register."""

    @property
    def end(self) -> int:
        """Index of the last qubit in the quantum register."""

    @property
    def size(self) -> int:
        """The number of qubits in the quantum register."""

    @property
    def name(self) -> str:
        """The name of the quantum register."""

    def __eq__(self, other: object) -> bool:
        """Check if the quantum register is equal to another quantum register."""

    def __ne__(self, other: object) -> bool:
        """Check if the quantum register is not equal to another quantum register."""

    def __hash__(self) -> int:
        """Return the hash of the quantum register."""

    def __getitem__(self, key: int) -> int:
        """Get the qubit at the specified index."""

    def __contains__(self, qubit: int) -> bool:
        """Check if the quantum register contains a qubit."""

class ClassicalRegister:
    """A class to represent a collection of classical bits.

    Args:
        start: The starting index of the classical register.
        size: The number of bits in the classical register.
        name: The name of the classical register. A name will be generated if not provided.
    """

    def __init__(self, start: int, size: int, name: str = "") -> None: ...
    @property
    def start(self) -> int:
        """The index of the first bit in the classical register."""

    @property
    def end(self) -> int:
        """Index of the last bit in the classical register."""

    @property
    def size(self) -> int:
        """The number of bits in the classical register."""

    @property
    def name(self) -> str:
        """The name of the classical register."""

    def __eq__(self, other: object) -> bool:
        """Check if the classical register is equal to another classical register."""

    def __ne__(self, other: object) -> bool:
        """Check if the classical register is not equal to another classical register."""

    def __hash__(self) -> int:
        """Return the hash of the classical register."""

    def __getitem__(self, key: int) -> int:
        """Get the bit at the specified index."""

    def __contains__(self, bit: int) -> bool:
        """Check if the classical register contains a bit."""



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""MQT Core IR  - The MQT Core Intermediate Representation (IR) module."""

from collections.abc import ItemsView, Iterable, Iterator, Mapping, MutableMapping, MutableSequence, Sequence
from os import PathLike
from typing import overload

from .operations import ComparisonKind, Control, Operation, OpType
from .registers import ClassicalRegister, QuantumRegister
from .symbolic import Expression, Variable

__all__ = [
    "Permutation",
    "QuantumComputation",
]

class Permutation(MutableMapping[int, int]):
    """A class to represent a permutation of the qubits in a quantum circuit.

    Args:
        permutation: The permutation to initialize the object with.

    """

    def __init__(self, permutation: dict[int, int] | None = None) -> None:
        """Initialize the permutation."""

    def __getitem__(self, idx: int) -> int:
        """Get the value of the permutation at the given index.

        Args:
            idx: The index to get the value of the permutation at.

        Returns:
            The value of the permutation at the given index.
        """

    def __setitem__(self, idx: int, val: int) -> None:
        """Set the value of the permutation at the given index.

        Args:
            idx: The index to set the value of the permutation at.
            val: The value to set the permutation at the given index to.
        """

    def __delitem__(self, key: int) -> None:
        """Delete the value of the permutation at the given index.

        Args:
            key: The index to delete the value of the permutation at.
        """

    def __iter__(self) -> Iterator[int]:
        """Return an iterator over the indices of the permutation."""

    def items(self) -> ItemsView[int, int]:
        """Return an iterable over the items of the permutation."""

    def __len__(self) -> int:
        """Return the number of indices in the permutation."""

    def __eq__(self, other: object) -> bool:
        """Check if the permutation is equal to another permutation."""

    def __ne__(self, other: object) -> bool:
        """Check if the permutation is not equal to another permutation."""

    def __hash__(self) -> int:
        """Return the hash of the permutation."""

    def clear(self) -> None:
        """Clear the permutation of all indices and values."""

    @overload
    def apply(self, controls: set[Control]) -> set[Control]:
        """Apply the permutation to a set of controls.

        Args:
            controls: The set of controls to apply the permutation to.

        Returns:
            The set of controls with the permutation applied.
        """

    @overload
    def apply(self, targets: list[int]) -> list[int]:
        """Apply the permutation to a list of targets.

        Args:
            targets: The list of targets to apply the permutation to.

        Returns:
            The list of targets with the permutation applied.
        """

class QuantumComputation(MutableSequence[Operation]):
    """The main class for representing quantum computations within the MQT.

    Acts as mutable sequence of :class:`~mqt.core.ir.operations.Operation` objects,
    which represent the individual operations in the quantum computation.

    Args:
        nq: The number of qubits in the quantum computation.
        nc: The number of classical bits in the quantum computation.
        seed: The seed to use for the internal random number generator.
    """

    # --------------------------------------------------------------------------
    #                               Constructors
    # --------------------------------------------------------------------------
    def __init__(self, nq: int = 0, nc: int = 0, seed: int = 0) -> None: ...
    @staticmethod
    def from_qasm_str(qasm: str) -> QuantumComputation:
        """Create a QuantumComputation object from an OpenQASM string.

        Args:
            qasm: The OpenQASM string to create the QuantumComputation object from.

        Returns:
            The QuantumComputation object created from the OpenQASM string.
        """
    @staticmethod
    def from_qasm(filename: str) -> QuantumComputation:
        """Create a QuantumComputation object from an OpenQASM file.

        Args:
            filename: The filename of the OpenQASM file to create the QuantumComputation object from.

        Returns:
            The QuantumComputation object created from the OpenQASM file.
        """

    # --------------------------------------------------------------------------
    #                          General Properties
    # --------------------------------------------------------------------------

    name: str
    """
    The name of the quantum computation.
    """
    global_phase: float
    """
    The global phase of the quantum computation.
    """

    @property
    def num_qubits(self) -> int:
        """The total number of qubits in the quantum computation."""

    @property
    def num_ancilla_qubits(self) -> int:
        r"""The number of ancilla qubits in the quantum computation.

        Note:
            Ancilla qubits are qubits that always start in a fixed state (usually :math:`|0\\rangle`).
        """

    @property
    def num_garbage_qubits(self) -> int:
        """The number of garbage qubits in the quantum computation.

        Note:
            Garbage qubits are qubits whose final state is not relevant for the computation.
        """

    @property
    def num_measured_qubits(self) -> int:
        """The number of qubits that are measured in the quantum computation.

        Computed as :math:`|qubits| - |garbage|`.
        """

    @property
    def num_data_qubits(self) -> int:
        """The number of data qubits in the quantum computation.

        Computed as :math:`|qubits| - |ancilla|`.
        """

    @property
    def num_classical_bits(self) -> int:
        """The number of classical bits in the quantum computation."""

    @property
    def num_ops(self) -> int:
        """The number of operations in the quantum computation."""

    def num_single_qubit_ops(self) -> int:
        """Return the number of single-qubit operations in the quantum computation."""

    def num_total_ops(self) -> int:
        """Return the total number of operations in the quantum computation.

        Recursively counts sub-operations (e.g., from :class:`~mqt.core.ir.operations.CompoundOperation` objects).
        """

    def depth(self) -> int:
        """Return the depth of the quantum computation."""

    def invert(self) -> None:
        """Invert the quantum computation in-place by inverting each operation and reversing the order of operations."""

    def to_operation(self) -> Operation:
        """Convert the quantum computation to a single operation.

        This gives ownership of the operations to the resulting operation,
        so the quantum computation will be empty after this operation.

        When the quantum computation contains more than one operation,
        the resulting operation is a :class:`~mqt.core.ir.operations.CompoundOperation`.

        Returns:
            The operation representing the quantum computation.
        """

    # --------------------------------------------------------------------------
    #                 Mutable Sequence Interface
    # --------------------------------------------------------------------------

    def __len__(self) -> int:
        """Return the number of operations in the quantum computation."""

    @overload
    def __getitem__(self, idx: int) -> Operation:
        """Get the operation at the given index.

        Note:
            This gives write access to the operation at the given index.

        Args:
            idx: The index of the operation to get.

        Returns:
            The operation at the given index.
        """

    @overload
    def __getitem__(self, idx: slice) -> list[Operation]:
        """Get a slice of operations from the quantum computation.

        Note:
            This gives write access to the operations in the given slice.

        Args:
            idx: The slice of operations to get.

        Returns:
            The operations in the given slice.
        """

    @overload
    def __setitem__(self, idx: int, op: Operation) -> None:
        """Set the operation at the given index.

        Args:
            idx: The index of the operation to set.
            op: The operation to set at the given index.
        """

    @overload
    def __setitem__(self, idx: slice, ops: Iterable[Operation]) -> None:
        """Set the operations in the given slice.

        Args:
            idx: The slice of operations to set.
            ops: The operations to set in the given slice.
        """

    @overload
    def __delitem__(self, idx: int) -> None:
        """Delete the operation at the given index.

        Args:
            idx: The index of the operation to delete.
        """

    @overload
    def __delitem__(self, idx: slice) -> None:
        """Delete the operations in the given slice.

        Args:
            idx: The slice of operations to delete.
        """

    def insert(self, idx: int, op: Operation) -> None:
        """Insert an operation at the given index.

        Args:
            idx: The index to insert the operation at.
            op: The operation to insert.
        """

    def append(self, op: Operation) -> None:
        """Append an operation to the end of the quantum computation.

        Args:
            op: The operation to append.
        """

    def reverse(self) -> None:
        """Reverse the order of the operations in the quantum computation (in-place)."""

    def clear(self) -> None:
        """Clear the quantum computation of all operations."""

    # --------------------------------------------------------------------------
    #                          (Qu)Bit Registers
    # --------------------------------------------------------------------------

    def add_ancillary_register(self, n: int, name: str = "anc") -> QuantumRegister:
        """Add an ancillary register to the quantum computation.

        Args:
            n: The number of qubits in the ancillary register.
            name: The name of the ancillary register.

        Returns:
            The ancillary register added to the quantum computation.
        """

    def add_classical_register(self, n: int, name: str = "c") -> ClassicalRegister:
        """Add a classical register to the quantum computation.

        Args:
            n: The number of bits in the classical register.
            name: The name of the classical register.

        Returns:
            The classical register added to the quantum computation.
        """

    def add_qubit_register(self, n: int, name: str = "q") -> QuantumRegister:
        """Add a qubit register to the quantum computation.

        Args:
            n: The number of qubits in the qubit register.
            name: The name of the qubit register.

        Returns:
            The qubit register added to the quantum computation.
        """

    def unify_quantum_registers(self, name: str = "q") -> QuantumRegister:
        """Unify all quantum registers in the quantum computation.

        Args:
            name: The name of the unified quantum register.

        Returns:
            The unified quantum register.
        """

    @property
    def qregs(self) -> dict[str, QuantumRegister]:
        """The quantum registers in the quantum computation."""

    @property
    def cregs(self) -> dict[str, ClassicalRegister]:
        """The classical registers in the quantum computation."""

    @property
    def ancregs(self) -> dict[str, QuantumRegister]:
        """The ancillary registers in the quantum computation."""

    # --------------------------------------------------------------------------
    #                  Initial Layout and Output Permutation
    # --------------------------------------------------------------------------

    initial_layout: Permutation
    """
    The initial layout of the qubits in the quantum computation.

    This is a permutation of the qubits in the quantum computation. It is mainly
    used to track the mapping of circuit qubits to device qubits during quantum
    circuit compilation. The keys are the device qubits (in which a compiled circuit
    is expressed in), and the values are the circuit qubits (in which the original
    quantum circuit is expressed in).

    Any operations in the quantum circuit are expected to be expressed in terms
    of the keys of the initial layout.

    Examples:
        - If no initial layout is explicitly specified (which is the default),
          the initial layout is assumed to be the identity permutation.
        - Assume a three-qubit circuit has been compiled to a four qubit device
          and circuit qubit 0 is mapped to device qubit 1, circuit qubit 1 is
          mapped to device qubit 2, and circuit qubit 2 is mapped to device qubit 3.
          Then the initial layout is {1: 0, 2: 1, 3: 2}.

    """
    output_permutation: Permutation
    """
    The output permutation of the qubits in the quantum computation.

    This is a permutation of the qubits in the quantum computation. It is mainly
    used to track where individual qubits end up at the end of the quantum computation,
    for example after a circuit has been compiled to a specific device and SWAP
    gates have been inserted, which permute the qubits. The keys are the qubits
    in the circuit and the values are the actual qubits being measured.

    Similar to the initial layout, the keys in the output permutation are the
    qubits actually present in the circuit and the values are the qubits in the
    "original" circuit.

    Examples:
        - If no output permutation is explicitly specified and the circuit does
          not contain measurements at the end, the output permutation is assumed
          to be the identity permutation.
        - If the circuit contains measurements at the end, these measurements
          are used to infer the output permutation. Assume a three-qubit circuit
          has been compiled to a four qubit device and, at the end of the circuit,
          circuit qubit 0 is measured into classical bit 2, circuit qubit 1 is
          measured into classical bit 1, and circuit qubit 3 is measured into
          classical bit 0. Then the output permutation is {0: 2, 1: 1, 3: 0}.
    """

    def initialize_io_mapping(self) -> None:
        """Initialize the I/O mapping of the quantum computation.

        If no initial layout is explicitly specified, the initial layout is assumed
        to be the identity permutation. If the circuit contains measurements at the
        end, these measurements are used to infer the output permutation.
        """

    # --------------------------------------------------------------------------
    #                       Ancilla and Garbage Handling
    # --------------------------------------------------------------------------

    @property
    def ancillary(self) -> list[bool]:
        """A list of booleans indicating whether each qubit is ancillary."""

    def set_circuit_qubit_ancillary(self, q: int) -> None:
        """Set a circuit (i.e., logical) qubit to be ancillary.

        Args:
            q: The index of the circuit qubit to set as ancillary.
        """

    def set_circuit_qubits_ancillary(self, q_min: int, q_max: int) -> None:
        """Set a range of circuit (i.e., logical) qubits to be ancillary.

        Args:
            q_min: The minimum index of the circuit qubits to set as ancillary.
            q_max: The maximum index of the circuit qubits to set as ancillary.
        """

    def is_circuit_qubit_ancillary(self, q: int) -> bool:
        """Check if a circuit (i.e., logical) qubit is ancillary.

        Args:
            q: The index of the circuit qubit to check.

        Returns:
            True if the circuit qubit is ancillary, False otherwise.
        """

    @property
    def garbage(self) -> list[bool]:
        """A list of booleans indicating whether each qubit is garbage."""

    def set_circuit_qubit_garbage(self, q: int) -> None:
        """Set a circuit (i.e., logical) qubit to be garbage.

        Args:
            q: The index of the circuit qubit to set as garbage.
        """

    def set_circuit_qubits_garbage(self, q_min: int, q_max: int) -> None:
        """Set a range of circuit (i.e., logical) qubits to be garbage.

        Args:
            q_min: The minimum index of the circuit qubits to set as garbage.
            q_max: The maximum index of the circuit qubits to set as garbage.
        """

    def is_circuit_qubit_garbage(self, q: int) -> bool:
        """Check if a circuit (i.e., logical) qubit is garbage.

        Args:
            q: The index of the circuit qubit to check.

        Returns:
            True if the circuit qubit is garbage, False otherwise.
        """

    # --------------------------------------------------------------------------
    #                        Symbolic Circuit Handling
    # --------------------------------------------------------------------------

    @property
    def variables(self) -> set[Variable]:
        """The set of variables in the quantum computation."""

    def add_variable(self, var: Expression | float) -> None:
        """Add a variable to the quantum computation.

        Args:
            var: The variable to add.
        """

    def add_variables(self, vars_: Sequence[Expression | float]) -> None:
        """Add multiple variables to the quantum computation.

        Args:
            vars_: The variables to add.
        """

    def is_variable_free(self) -> bool:
        """Check if the quantum computation is free of variables.

        Returns:
            True if the quantum computation is free of variables, False otherwise.
        """

    def instantiate(self, assignment: Mapping[Variable, float]) -> QuantumComputation:
        """Instantiate the quantum computation with the given variable assignment.

        Args:
            assignment: The variable assignment to instantiate the quantum computation with.

        Returns:
            The instantiated quantum computation.
        """

    def instantiate_inplace(self, assignment: Mapping[Variable, float]) -> None:
        """Instantiate the quantum computation with the given variable assignment in-place.

        Args:
            assignment: The variable assignment to instantiate the quantum computation with.
        """

    # --------------------------------------------------------------------------
    #                             Output Handling
    # --------------------------------------------------------------------------

    def qasm2_str(self) -> str:
        """Return the OpenQASM2 representation of the quantum computation as a string.

        Note:
            This uses some custom extensions to OpenQASM 2.0 that allow for easier
            definition of multi-controlled gates. These extensions might not be
            supported by all OpenQASM 2.0 parsers. Consider using the :meth:`qasm3_str`
            method instead, which uses OpenQASM 3.0 that natively supports
            multi-controlled gates. The export also assumes the bigger, non-standard
            `qelib1.inc` from Qiskit is available.

        Returns:
            The OpenQASM2 representation of the quantum computation as a string.
        """

    def qasm2(self, filename: PathLike[str] | str) -> None:
        """Write the OpenQASM2 representation of the quantum computation to a file.

        See Also:
            :meth:`qasm2_str`

        Args:
            filename: The filename of the file to write the OpenQASM2 representation to.
        """

    def qasm3_str(self) -> str:
        """Return the OpenQASM3 representation of the quantum computation as a string.

        Returns:
            The OpenQASM3 representation of the quantum computation as a string.
        """

    def qasm3(self, filename: PathLike[str] | str) -> None:
        """Write the OpenQASM3 representation of the quantum computation to a file.

        See Also:
            :meth:`qasm3_str`

        Args:
            filename: The filename of the file to write the OpenQASM3 representation to.
        """

    # --------------------------------------------------------------------------
    #                               Operations
    # --------------------------------------------------------------------------

    def i(self, q: int) -> None:
        r"""Apply an identity operation.

        .. math::
            I = \begin{pmatrix} 1 & 0 \\ 0 & 1 \end{pmatrix}

        Args:
            q: The target qubit
        """

    def ci(self, control: Control | int, target: int) -> None:
        """Apply a controlled identity operation.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`i`
        """

    def mci(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled identity operation.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`i`
        """

    def x(self, q: int) -> None:
        r"""Apply a Pauli-X gate.

        .. math::
            X = \begin{pmatrix} 0 & 1 \\ 1 & 0 \end{pmatrix}

        Args:
            q: The target qubit
        """

    def cx(self, control: Control | int, target: int) -> None:
        """Apply a controlled Pauli-X (CNOT / CX) gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`x`
        """

    def mcx(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled Pauli-X (Toffoli / MCX) gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`x`
        """

    def y(self, q: int) -> None:
        r"""Apply a Pauli-Y gate.

        .. math::
            Y = \begin{pmatrix} 0 & -i \\ i & 0 \end{pmatrix}

        Args:
            q: The target qubit
        """

    def cy(self, control: Control | int, target: int) -> None:
        """Apply a controlled Pauli-Y gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`y`
        """

    def mcy(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled Pauli-Y gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`y`
        """

    def z(self, q: int) -> None:
        r"""Apply a Pauli-Z gate.

        .. math::
            Z = \begin{pmatrix} 1 & 0 \\ 0 & -1 \end{pmatrix}

        Args:
            q: The target qubit
        """

    def cz(self, control: Control | int, target: int) -> None:
        """Apply a controlled Pauli-Z (CZ) gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`z`
        """

    def mcz(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled Pauli-Z (MCZ) gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`z`
        """

    def h(self, q: int) -> None:
        r"""Apply a Hadamard gate.

        .. math::
            H = \frac{1}{\sqrt{2}} \begin{pmatrix} 1 & 1 \\ 1 & -1 \end{pmatrix}

        Args:
            q: The target qubit
        """

    def ch(self, control: Control | int, target: int) -> None:
        """Apply a controlled Hadamard gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`h`
        """

    def mch(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled Hadamard gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`h`
        """

    def s(self, q: int) -> None:
        r"""Apply an S gate (phase gate).

        .. math::
            S = \begin{pmatrix} 1 & 0 \\ 0 & i \end{pmatrix}

        Args:
            q: The target qubit
        """

    def cs(self, control: Control | int, target: int) -> None:
        """Apply a controlled S gate (CS gate).

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`s`
        """

    def mcs(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled S gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`s`
        """

    def sdg(self, q: int) -> None:
        r"""Apply an :math:`S^{\dagger}` gate.

        .. math::
            S^{\dagger} = \begin{pmatrix} 1 & 0 \\ 0 & -i \end{pmatrix}

        Args:
            q: The target qubit
        """

    def csdg(self, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`S^{\dagger}` gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`sdg`
        """

    def mcsdg(self, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`S^{\dagger}` gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`sdg`
        """

    def t(self, q: int) -> None:
        r"""Apply a T gate.

        .. math::
            T = \begin{pmatrix} 1 & 0 \\ 0 & e^{i\pi/4} \end{pmatrix}

        Args:
            q: The target qubit
        """

    def ct(self, control: Control | int, target: int) -> None:
        """Apply a controlled T gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`t`
        """

    def mct(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled T gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`t`
        """

    def tdg(self, q: int) -> None:
        r"""Apply a :math:`T^{\dagger}` gate.

        .. math::
            T^{\dagger} = \begin{pmatrix} 1 & 0 \\ 0 & e^{-i\pi/4} \end{pmatrix}

        Args:
            q: The target qubit
        """

    def ctdg(self, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`T^{\dagger}` gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`tdg`
        """

    def mctdg(self, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`T^{\dagger}` gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`tdg`
        """

    def v(self, q: int) -> None:
        r"""Apply a V gate.

        .. math::
            V = \frac{1}{\sqrt{2}} \begin{pmatrix} 1 & -i \\ -i & 1 \end{pmatrix}

        Args:
            q: The target qubit
        """

    def cv(self, control: Control | int, target: int) -> None:
        """Apply a controlled V gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`v`
        """

    def mcv(self, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled V gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`v`
        """

    def vdg(self, q: int) -> None:
        r"""Apply a :math:`V^{\dagger}` gate.

        .. math::
            V^{\dagger} = \frac{1}{\sqrt{2}} \begin{pmatrix} 1 & i \\ i & 1 \end{pmatrix}

        Args:
            q: The target qubit
        """

    def cvdg(self, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`V^{\dagger}` gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`vdg`
        """

    def mcvdg(self, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`V^{\dagger}` gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`vdg`
        """

    def sx(self, q: int) -> None:
        r"""Apply a :math:`\sqrt{X}` gate.

        .. math::
            \sqrt{X} = \frac{1}{2} \begin{pmatrix} 1 + i & 1 - i \\ 1 - i & 1 + i \end{pmatrix}

        Args:
            q: The target qubit
        """

    def csx(self, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`\sqrt{X}` gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`sx`
        """

    def mcsx(self, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`\sqrt{X}` gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`sx`
        """

    def sxdg(self, q: int) -> None:
        r"""Apply a :math:`\sqrt{X}^{\dagger}` gate.

        .. math::
            \sqrt{X}^{\dagger} = \frac{1}{2} \begin{pmatrix} 1 - i & 1 + i \\ 1 + i & 1 - i \end{pmatrix}

        Args:
            q: The target qubit
        """

    def csxdg(self, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`\sqrt{X}^{\dagger}` gate.

        Args:
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`sxdg`
        """

    def mcsxdg(self, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`\sqrt{X}^{\dagger}` gate.

        Args:
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`sxdg`
        """

    def r(self, theta: float | Expression, phi: float | Expression, q: int) -> None:
        r"""Apply an :math:`R(\theta, \phi)` gate.

        .. math::
            R(\theta, \phi) = e^{-i\frac{\theta}{2}(\cos(\phi)X+\sin(\phi)Y)}
            = \begin{pmatrix} \cos(\theta/2) & -i e^{-i\phi} \sin(\theta/2) \\
            -i e^{i\phi} \sin(\theta/2) & \cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle :math:`\theta`
            phi: The angle specifying the rotation axis given by :math:`\cos(\phi)X+\sin(\phi)Y`
            q: The target qubit
        """

    def cr(self, theta: float | Expression, phi: float | Expression, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`R(\theta, \phi)` gate.

        Args:
            theta: The rotation angle :math:`\theta`
            phi: The angle specifying the rotation axis given by :math:`\cos(\phi)X+\sin(\phi)Y`
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`r`
        """

    def mcr(
        self, theta: float | Expression, phi: float | Expression, controls: set[Control | int], target: int
    ) -> None:
        r"""Apply a multi-controlled :math:`R(\theta, \phi)` gate.

        Args:
            theta: The rotation angle :math:`\theta`
            phi: The angle specifying the rotation axis given by :math:`\cos(\phi)X+\sin(\phi)Y`
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`r`
        """

    def rx(self, theta: float | Expression, q: int) -> None:
        r"""Apply an :math:`R_x(\theta)` gate.

        .. math::
            R_x(\theta) = e^{-i\theta X/2} = \cos(\theta/2) I - i \sin(\theta/2) X
            = \begin{pmatrix} \cos(\theta/2) & -i \sin(\theta/2) \\ -i \sin(\theta/2) & \cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle
            q: The target qubit
        """

    def crx(self, theta: float | Expression, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`R_x(\theta)` gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`rx`
        """

    def mcrx(self, theta: float | Expression, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`R_x(\theta)` gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`rx`
        """

    def ry(self, theta: float | Expression, q: int) -> None:
        r"""Apply an :math:`R_y(\theta)` gate.

        .. math::
            R_y(\theta) = e^{-i\theta Y/2} = \cos(\theta/2) I - i \sin(\theta/2) Y
            = \begin{pmatrix} \cos(\theta/2) & -\sin(\theta/2) \\ \sin(\theta/2) & \cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle
            q: The target qubit
        """

    def cry(self, theta: float | Expression, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`R_y(\theta)` gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`ry`
        """

    def mcry(self, theta: float | Expression, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`R_y(\theta)` gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`ry`
        """

    def rz(self, theta: float | Expression, q: int) -> None:
        r"""Apply an :math:`R_z(\theta)` gate.

        .. math::
            R_z(\theta) = e^{-i\theta Z/2} = \begin{pmatrix} e^{-i\theta/2} & 0 \\ 0 & e^{i\theta/2} \end{pmatrix}

        Args:
            theta: The rotation angle
            q: The target qubit
        """

    def crz(self, theta: float | Expression, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`R_z(\theta)` gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`rz`
        """

    def mcrz(self, theta: float | Expression, controls: set[Control | int], target: int) -> None:
        r"""Apply a multi-controlled :math:`R_z(\theta)` gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`rz`
        """

    def p(self, theta: float | Expression, q: int) -> None:
        r"""Apply a phase gate.

        .. math::
            P(\theta) = \begin{pmatrix} 1 & 0 \\ 0 & e^{i\theta} \end{pmatrix}

        Args:
            theta: The rotation angle
            q: The target qubit
        """

    def cp(self, theta: float | Expression, control: Control | int, target: int) -> None:
        """Apply a controlled phase gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`p`
        """

    def mcp(self, theta: float | Expression, controls: set[Control | int], target: int) -> None:
        """Apply a multi-controlled phase gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`p`
        """

    def u2(self, phi: float | Expression, lambda_: float | Expression, q: int) -> None:
        r"""Apply a :math:`U_2(\phi, \lambda)` gate.

        .. math::
            U_2(\phi, \lambda) =
            \frac{1}{\sqrt{2}} \begin{pmatrix} 1 & -e^{i\lambda} \\ e^{i\phi} & e^{i(\phi + \lambda)} \end{pmatrix}

        Args:
            phi: The rotation angle
            lambda_: The rotation angle
            q: The target qubit
        """

    def cu2(self, phi: float | Expression, lambda_: float | Expression, control: Control | int, target: int) -> None:
        r"""Apply a controlled :math:`U_2(\phi, \lambda)` gate.

        Args:
            phi: The rotation angle
            lambda_: The rotation angle
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`u2`
        """

    def mcu2(
        self,
        phi: float | Expression,
        lambda_: float | Expression,
        controls: set[Control | int],
        target: int,
    ) -> None:
        r"""Apply a multi-controlled :math:`U_2(\phi, \lambda)` gate.

        Args:
            phi: The rotation angle
            lambda_: The rotation angle
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`u2`
        """

    def u(self, theta: float | Expression, phi: float | Expression, lambda_: float | Expression, q: int) -> None:
        r"""Apply a :math:`U(\theta, \phi, \lambda)` gate.

        .. math::
            U(\theta, \phi, \lambda) =
            \begin{pmatrix} \cos(\theta/2) & -e^{i\lambda}\sin(\theta/2) \\
            e^{i\phi}\sin(\theta/2) & e^{i(\phi + \lambda)}\cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle
            phi: The rotation angle
            lambda_: The rotation angle
            q: The target qubit
        """

    def cu(
        self,
        theta: float | Expression,
        phi: float | Expression,
        lambda_: float | Expression,
        control: Control | int,
        target: int,
    ) -> None:
        r"""Apply a controlled :math:`U(\theta, \phi, \lambda)` gate.

        Args:
            theta: The rotation angle
            phi: The rotation angle
            lambda_: The rotation angle
            control: The control qubit
            target: The target qubit

        See Also:
            :meth:`u`
        """

    def mcu(
        self,
        theta: float | Expression,
        phi: float | Expression,
        lambda_: float | Expression,
        controls: set[Control | int],
        target: int,
    ) -> None:
        r"""Apply a multi-controlled :math:`U(\theta, \phi, \lambda)` gate.

        Args:
            theta: The rotation angle
            phi: The rotation angle
            lambda_: The rotation angle
            controls: The control qubits
            target: The target qubit

        See Also:
            :meth:`u`
        """

    def swap(self, target1: int, target2: int) -> None:
        r"""Apply a SWAP gate.

        .. math::
            SWAP = \begin{pmatrix} 1 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1 \end{pmatrix}

        Args:
            target1: The first target qubit
            target2: The second target qubit
        """

    def cswap(self, control: Control | int, target1: int, target2: int) -> None:
        """Apply a controlled SWAP gate.

        Args:
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`swap`
        """

    def mcswap(self, controls: set[Control | int], target1: int, target2: int) -> None:
        """Apply a multi-controlled SWAP gate.

        Args:
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`swap`
        """

    def dcx(self, target1: int, target2: int) -> None:
        r"""Apply a DCX (double CNOT) gate.

        .. math::
            DCX = \begin{pmatrix} 1 & 0 & 0 & 0 \\ 0 & 0 & 0 & 1 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 1 & 0 \end{pmatrix}

        Args:
            target1: The first target qubit
            target2: The second target qubit
        """

    def cdcx(self, control: Control | int, target1: int, target2: int) -> None:
        """Apply a controlled DCX (double CNOT) gate.

        Args:
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`dcx`
        """

    def mcdcx(self, controls: set[Control | int], target1: int, target2: int) -> None:
        """Apply a multi-controlled DCX (double CNOT) gate.

        Args:
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`dcx`
        """

    def ecr(self, target1: int, target2: int) -> None:
        r"""Apply an ECR (echoed cross-resonance) gate.

        .. math::
            ECR = \frac{1}{\sqrt{2}}
            \begin{pmatrix} 0 & 0 & 1 & i \\ 0 & 0 & i & 1 \\ 1 & -i & 0 & 0 \\ -i & 1 & 0 & 0 \end{pmatrix}

        Args:
            target1: The first target qubit
            target2: The second target qubit
        """

    def cecr(self, control: Control | int, target1: int, target2: int) -> None:
        """Apply a controlled ECR (echoed cross-resonance) gate.

        Args:
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`ecr`
        """

    def mcecr(self, controls: set[Control | int], target1: int, target2: int) -> None:
        """Apply a multi-controlled ECR (echoed cross-resonance) gate.

        Args:
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`ecr`
        """

    def iswap(self, target1: int, target2: int) -> None:
        r"""Apply an iSWAP gate.

        .. math::
            iSWAP = \begin{pmatrix} 1 & 0 & 0 & 0 \\ 0 & 0 & i & 0 \\ 0 & i & 0 & 0 \\ 0 & 0 & 0 & 1 \end{pmatrix}

        Args:
            target1: The first target qubit
            target2: The second target qubit
        """

    def ciswap(self, control: Control | int, target1: int, target2: int) -> None:
        """Apply a controlled iSWAP gate.

        Args:
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`iswap`
        """

    def mciswap(self, controls: set[Control | int], target1: int, target2: int) -> None:
        """Apply a multi-controlled iSWAP gate.

        Args:
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`iswap`
        """

    def iswapdg(self, target1: int, target2: int) -> None:
        r"""Apply an :math:`iSWAP^{\dagger}` gate.

        .. math::
            iSWAP^{\dagger} =
            \begin{pmatrix} 1 & 0 & 0 & 0 \\ 0 & 0 & -i & 0 \\ 0 & -i & 0 & 0 \\ 0 & 0 & 0 & 1 \end{pmatrix}

        Args:
            target1: The first target qubit
            target2: The second target qubit
        """

    def ciswapdg(self, control: Control | int, target1: int, target2: int) -> None:
        r"""Apply a controlled :math:`iSWAP^{\dagger}` gate.

        Args:
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`iswapdg`
        """

    def mciswapdg(self, controls: set[Control | int], target1: int, target2: int) -> None:
        r"""Apply a multi-controlled :math:`iSWAP^{\dagger}` gate.

        Args:
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`iswapdg`
        """

    def peres(self, target1: int, target2: int) -> None:
        r"""Apply a Peres gate.

        .. math::
            Peres = \begin{pmatrix} 0 & 0 & 0 & 1 \\ 0 & 0 & 1 & 0 \\ 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 \end{pmatrix}

        Args:
            target1: The first target qubit
            target2: The second target qubit
        """

    def cperes(self, control: Control | int, target1: int, target2: int) -> None:
        """Apply a controlled Peres gate.

        Args:
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`peres`
        """

    def mcperes(self, controls: set[Control | int], target1: int, target2: int) -> None:
        """Apply a multi-controlled Peres gate.

        Args:
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`peres`
        """

    def peresdg(self, target1: int, target2: int) -> None:
        r"""Apply a :math:`Peres^{\dagger}` gate.

        .. math::
            Peres^{\dagger} =
            \begin{pmatrix} 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 1 \\ 0 & 1 & 0 & 0 \\ 1 & 0 & 0 & 0 \end{pmatrix}

        Args:
            target1: The first target qubit
            target2: The second target qubit
        """

    def cperesdg(self, control: Control | int, target1: int, target2: int) -> None:
        r"""Apply a controlled :math:`Peres^{\dagger}` gate.

        Args:
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`peresdg`
        """

    def mcperesdg(self, controls: set[Control | int], target1: int, target2: int) -> None:
        r"""Apply a multi-controlled :math:`Peres^{\dagger}` gate.

        Args:
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`peresdg`
        """

    def rxx(self, theta: float | Expression, target1: int, target2: int) -> None:
        r"""Apply an :math:`R_{xx}(\theta)` gate.

        .. math::
            R_{xx}(\theta) = e^{-i\theta XX/2} = \cos(\theta/2) I\otimes I - i \sin(\theta/2) X \otimes X
            = \begin{pmatrix} \cos(\theta/2) & 0 & 0 & -i \sin(\theta/2) \\
            0 & \cos(\theta/2) & -i \sin(\theta/2) & 0 \\
            0 & -i \sin(\theta/2) & \cos(\theta/2) & 0 \\
            -i \sin(\theta/2) & 0 & 0 & \cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle
            target1: The first target qubit
            target2: The second target qubit
        """

    def crxx(self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None:
        r"""Apply a controlled :math:`R_{xx}(\theta)` gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`rxx`
        """

    def mcrxx(self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int) -> None:
        r"""Apply a multi-controlled :math:`R_{xx}(\theta)` gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`rxx`
        """

    def ryy(self, theta: float | Expression, target1: int, target2: int) -> None:
        r"""Apply an :math:`R_{yy}(\theta)` gate.

        .. math::
            R_{yy}(\theta) = e^{-i\theta YY/2} = \cos(\theta/2) I\otimes I - i \sin(\theta/2) Y \otimes Y
            = \begin{pmatrix} \cos(\theta/2) & 0 & 0 & i \sin(\theta/2) \\
            0 & \cos(\theta/2) & -i \sin(\theta/2) & 0 \\
            0 & -i \sin(\theta/2) & \cos(\theta/2) & 0 \\
            i \sin(\theta/2) & 0 & 0 & \cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle
            target1: The first target qubit
            target2: The second target qubit
        """

    def cryy(self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None:
        r"""Apply a controlled :math:`R_{yy}(\theta)` gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`ryy`
        """

    def mcryy(self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int) -> None:
        r"""Apply a multi-controlled :math:`R_{yy}(\theta)` gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`ryy`
        """

    def rzz(self, theta: float | Expression, target1: int, target2: int) -> None:
        r"""Apply an :math:`R_{zz}(\theta)` gate.

        .. math::
            R_{zz}(\theta) = e^{-i\theta ZZ/2} =
            \begin{pmatrix} e^{-i\theta/2} & 0 & 0 & 0 \\
            0 & e^{i\theta/2} & 0 & 0 \\
            0 & 0 & e^{i\theta/2} & 0 \\
            0 & 0 & 0 & e^{-i\theta/2} \end{pmatrix}

        Args:
            theta: The rotation angle
            target1: The first target qubit
            target2: The second target qubit
        """

    def crzz(self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None:
        r"""Apply a controlled :math:`R_{zz}(\theta)` gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`rzz`
        """

    def mcrzz(self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int) -> None:
        r"""Apply a multi-controlled :math:`R_{zz}(\theta)` gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`rzz`
        """

    def rzx(self, theta: float | Expression, target1: int, target2: int) -> None:
        r"""Apply an :math:`R_{zx}(\theta)` gate.

        .. math::
            R_{zx}(\theta) = e^{-i\theta ZX/2} = \cos(\theta/2) I\otimes I - i \sin(\theta/2) Z \otimes X =
            \begin{pmatrix} \cos(\theta/2) & -i \sin(\theta/2) & 0 & 0 \\
            -i \sin(\theta/2) & \cos(\theta/2) & 0 & 0 \\
            0 & 0 & \cos(\theta/2) & i \sin(\theta/2) \\
            0 & 0 & i \sin(\theta/2) & \cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle
            target1: The first target qubit
            target2: The second target qubit
        """

    def crzx(self, theta: float | Expression, control: Control | int, target1: int, target2: int) -> None:
        r"""Apply a controlled :math:`R_{zx}(\theta)` gate.

        Args:
            theta: The rotation angle
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`rzx`
        """

    def mcrzx(self, theta: float | Expression, controls: set[Control | int], target1: int, target2: int) -> None:
        r"""Apply a multi-controlled :math:`R_{zx}(\theta)` gate.

        Args:
            theta: The rotation angle
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`rzx`
        """

    def xx_minus_yy(self, theta: float | Expression, beta: float | Expression, target1: int, target2: int) -> None:
        r"""Apply an :math:`R_{XX - YY}(\theta, \beta)` gate.

        .. math::
            R_{XX - YY}(\theta, \beta)
            = R_{z_2}(\beta) \cdot e^{-i\frac{\theta}{2} \frac{XX-YY}{2}} \cdot R_{z_2}(-\beta)
            = \begin{pmatrix} \cos(\theta/2) & 0 & 0 & -i \sin(\theta/2) e^{-i\beta} \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            -i \sin(\theta/2) e^{i\beta} & 0 & 0 & \cos(\theta/2) \end{pmatrix}

        Args:
            theta: The rotation angle
            beta: The rotation angle
            target1: The first target qubit
            target2: The second target qubit
        """

    def cxx_minus_yy(
        self,
        theta: float | Expression,
        beta: float | Expression,
        control: Control | int,
        target1: int,
        target2: int,
    ) -> None:
        r"""Apply a controlled :math:`R_{XX - YY}(\theta, \beta)` gate.

        Args:
            theta: The rotation angle
            beta: The rotation angle
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`xx_minus_yy`
        """

    def mcxx_minus_yy(
        self,
        theta: float | Expression,
        beta: float | Expression,
        controls: set[Control | int],
        target1: int,
        target2: int,
    ) -> None:
        r"""Apply a multi-controlled :math:`R_{XX - YY}(\theta, \beta)` gate.

        Args:
            theta: The rotation angle
            beta: The rotation angle
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`xx_minus_yy`
        """

    def xx_plus_yy(self, theta: float | Expression, beta: float | Expression, target1: int, target2: int) -> None:
        r"""Apply an :math:`R_{XX + YY}(\theta, \beta)` gate.

        .. math::
            R_{XX + YY}(\theta, \beta)
            = R_{z_1}(\beta) \cdot e^{-i\frac{\theta}{2} \frac{XX+YY}{2}} \cdot R_{z_1}(-\beta)
            = \begin{pmatrix} 1 & 0 & 0 & 0 \\
            0 & \cos(\theta/2) & -i \sin(\theta/2) e^{-i\beta} & 0 \\
            0 & -i \sin(\theta/2) e^{i\beta} & \cos(\theta/2) & 0 \\
            0 & 0 & 0 & 1 \end{pmatrix}

        Args:
            theta: The rotation angle
            beta: The rotation angle
            target1: The first target qubit
            target2: The second target qubit
        """

    def cxx_plus_yy(
        self,
        theta: float | Expression,
        beta: float | Expression,
        control: Control | int,
        target1: int,
        target2: int,
    ) -> None:
        r"""Apply a controlled :math:`R_{XX + YY}(\theta, \beta)` gate.

        Args:
            theta: The rotation angle
            beta: The rotation angle
            control: The control qubit
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`xx_plus_yy`
        """

    def mcxx_plus_yy(
        self,
        theta: float | Expression,
        beta: float | Expression,
        controls: set[Control | int],
        target1: int,
        target2: int,
    ) -> None:
        r"""Apply a multi-controlled :math:`R_{XX + YY}(\theta, \beta)` gate.

        Args:
            theta: The rotation angle
            beta: The rotation angle
            controls: The control qubits
            target1: The first target qubit
            target2: The second target qubit

        See Also:
            :meth:`xx_plus_yy`
        """

    def gphase(self, theta: float) -> None:
        r"""Apply a global phase gate.

        .. math::
            GPhase(\theta) = (e^{i\theta})

        Args:
            theta: The rotation angle
        """

    @overload
    def measure(self, qubit: int, cbit: int) -> None:
        """Measure a qubit and store the result in a classical bit.

        Args:
            qubit: The qubit to measure
            cbit: The classical bit to store the result
        """

    @overload
    def measure(self, qubits: Sequence[int], cbits: Sequence[int]) -> None:
        """Measure multiple qubits and store the results in classical bits.

        This method is equivalent to calling :meth:`measure` multiple times.

        Args:
            qubits: The qubits to measure
            cbits: The classical bits to store the results
        """

    def measure_all(self, *, add_bits: bool = True) -> None:
        """Measure all qubits and store the results in classical bits.

        Details:
            If `add_bits` is `True`, a new classical register (named "`meas`") with
            the same size as the number of qubits will be added to the circuit
            and the results will be stored in it. If `add_bits` is `False`, the
            classical register must already exist and have a sufficient number
            of bits to store the results.

        Args:
            add_bits: Whether to explicitly add a classical register
        """

    @overload
    def reset(self, q: int) -> None:
        """Add a reset operation to the circuit.

        Args:
            q: The qubit to reset
        """

    @overload
    def reset(self, qubits: Sequence[int]) -> None:
        """Add a reset operation to the circuit.

        Args:
            qubits: The qubits to reset
        """

    @overload
    def barrier(self) -> None:
        """Add a barrier to the circuit."""

    @overload
    def barrier(self, q: int) -> None:
        """Add a barrier to the circuit.

        Args:
            q: The qubit to add the barrier to
        """

    @overload
    def barrier(self, qubits: Sequence[int]) -> None:
        """Add a barrier to the circuit.

        Args:
            qubits: The qubits to add the barrier to
        """

    @overload
    def if_else(
        self,
        then_operation: Operation,
        else_operation: Operation | None,
        control_register: ClassicalRegister,
        expected_value: int = 1,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
    ) -> None:
        """Add an if-else operation to the circuit.

        Args:
            then_operation: The operation to apply if the condition is met
            else_operation: The operation to apply if the condition is not met
            control_register: The classical register to check against
            expected_value: The expected value of the control register
            comparison_kind: The kind of comparison to perform
        """

    @overload
    def if_else(
        self,
        then_operation: Operation,
        else_operation: Operation | None,
        control_bit: int,
        expected_value: int = 1,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
    ) -> None:
        """Add an if-else operation to the circuit.

        Args:
            then_operation: The operation to apply if the condition is met
            else_operation: The operation to apply if the condition is not met
            control_bit: The index of the classical bit to check against
            expected_value: The expected value of the control bit
            comparison_kind: The kind of comparison to perform
        """

    @overload
    def if_(
        self,
        op_type: OpType,
        target: int,
        control_register: ClassicalRegister,
        expected_value: int = 1,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
        params: Sequence[float] = (),
    ) -> None:
        """Add an if operartion to the circuit.

        Args:
            op_type: The operation to apply
            target: The target qubit
            control_register: The classical register to check against
            expected_value: The expected value of the control register
            comparison_kind: The kind of comparison to perform
            params: The parameters of the operation
        """

    @overload
    def if_(
        self,
        op_type: OpType,
        target: int,
        control: Control | int,
        control_register: ClassicalRegister,
        expected_value: int = 1,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
        params: Sequence[float] = (),
    ) -> None:
        """Add a classic-controlled operation to the circuit.

        Args:
            op_type: The operation to apply
            target: The target qubit
            control: The control qubit
            control_register: The classical register to check against
            expected_value: The expected value of the control register
            comparison_kind: The kind of comparison to perform
            params: The parameters of the operation.
        """

    @overload
    def if_(
        self,
        op_type: OpType,
        target: int,
        controls: set[Control | int],
        control_register: ClassicalRegister,
        expected_value: int = 1,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
        params: Sequence[float] = (),
    ) -> None:
        """Add a classic-controlled operation to the circuit.

        Args:
            op_type: The operation to apply
            target: The target qubit
            controls: The control qubits
            control_register: The classical register to check against
            expected_value: The expected value of the control register
            comparison_kind: The kind of comparison to perform
            params: The parameters of the operation.
        """

    @overload
    def if_(
        self,
        op_type: OpType,
        target: int,
        control_bit: int,
        expected_value: bool = True,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
        params: Sequence[float] = (),
    ) -> None:
        """Add a classic-controlled operation to the circuit.

        Args:
            op_type: The operation to apply
            target: The target qubit
            control_bit: The index of the classical bit to check against
            expected_value: The expected value of the control bit
            comparison_kind: The kind of comparison to perform
            params: The parameters of the operation.
        """

    @overload
    def if_(
        self,
        op_type: OpType,
        target: int,
        control: Control | int,
        control_bit: int,
        expected_value: bool = True,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
        params: Sequence[float] = (),
    ) -> None:
        """Add a classic-controlled operation to the circuit.

        Args:
            op_type: The operation to apply
            target: The target qubit
            control: The control qubit
            control_bit: The index of the classical bit to check against
            expected_value: The expected value of the control bit
            comparison_kind: The kind of comparison to perform
            params: The parameters of the operation.
        """

    @overload
    def if_(
        self,
        op_type: OpType,
        target: int,
        controls: set[Control | int],
        control_bit: int,
        expected_value: bool = True,
        comparison_kind: ComparisonKind = ComparisonKind.eq,
        params: Sequence[float] = (),
    ) -> None:
        """Add a classic-controlled operation to the circuit.

        Args:
            op_type: The operation to apply
            target: The target qubit
            controls: The control qubits
            control_bit: The index of the classical bit to check against
            expected_value: The expected value of the control bit
            comparison_kind: The kind of comparison to perform
            params: The parameters of the operation.
        """



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""MQT Core - The Backbone of the Munich Quantum Toolkit."""

from __future__ import annotations

import os
import sys
from pathlib import Path

# under Windows, make sure to add the appropriate DLL directory to the PATH
if sys.platform == "win32":

    def _dll_patch() -> None:
        """Add the DLL directory to the PATH."""
        import sysconfig

        bin_dir = Path(sysconfig.get_paths()["purelib"]) / "mqt" / "core" / "bin"
        os.add_dll_directory(str(bin_dir))

    _dll_patch()
    del _dll_patch

from typing import TYPE_CHECKING

from ._version import version as __version__
from ._version import version_tuple as version_info
from .ir import QuantumComputation

if TYPE_CHECKING:
    from qiskit.circuit import QuantumCircuit


def load(input_circuit: QuantumComputation | str | os.PathLike[str] | QuantumCircuit) -> QuantumComputation:
    """Load a quantum circuit from any supported format as a :class:`~mqt.core.ir.QuantumComputation`.

    Args:
        input_circuit: The input circuit to translate to a :class:`~mqt.core.ir.QuantumComputation`.
                       This can be a :class:`~mqt.core.ir.QuantumComputation` itself,
                       a file name to any of the supported file formats,
                       an OpenQASM (2.0 or 3.0) string, or
                       a Qiskit :class:`~qiskit.circuit.QuantumCircuit`.

    Returns:
        The :class:`~mqt.core.ir.QuantumComputation`.

    Raises:
        FileNotFoundError: If the input circuit is a file name and the file does not exist.
    """
    if isinstance(input_circuit, QuantumComputation):
        return input_circuit

    if isinstance(input_circuit, (str, os.PathLike)):
        input_str = str(input_circuit)
        max_filename_length = 255 if os.name == "nt" else os.pathconf("/", "PC_NAME_MAX")
        if len(input_str) > max_filename_length or not Path(input_circuit).is_file():
            if isinstance(input_circuit, os.PathLike) or "OPENQASM" not in input_circuit:
                msg = f"File {input_circuit} does not exist."
                raise FileNotFoundError(msg)
            # otherwise, we assume that this is a QASM string
            return QuantumComputation.from_qasm_str(input_str)

        return QuantumComputation.from_qasm(input_str)

    # At this point, we know that the input is a Qiskit QuantumCircuit
    from .plugins.qiskit import qiskit_to_mqt

    return qiskit_to_mqt(input_circuit)


__all__ = ["__version__", "load", "version_info"]



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Useful commands for obtaining information about mqt-core."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, distribution
from pathlib import Path


def include_dir() -> Path:
    """Return the path to the mqt-core include directory.

    Raises:
        FileNotFoundError: If the include directory is not found.
        ImportError: If mqt-core is not installed.
    """
    try:
        dist = distribution("mqt-core")
        located_include_dir = Path(str(dist.locate_file("mqt/core/include/mqt-core")))
        if located_include_dir.exists() and located_include_dir.is_dir():
            return located_include_dir
        msg = "mqt-core include files not found."
        raise FileNotFoundError(msg)
    except PackageNotFoundError:
        msg = "mqt-core not installed, installation required to access the include files."
        raise ImportError(msg) from None


def cmake_dir() -> Path:
    """Return the path to the mqt-core CMake module directory.

    Raises:
        FileNotFoundError: If the CMake module directory is not found.
        ImportError: If mqt-core is not installed.
    """
    try:
        dist = distribution("mqt-core")
        located_cmake_dir = Path(str(dist.locate_file("mqt/core/share/cmake")))
        if located_cmake_dir.exists() and located_cmake_dir.is_dir():
            return located_cmake_dir
        msg = "mqt-core CMake files not found."
        raise FileNotFoundError(msg)
    except PackageNotFoundError:
        msg = "mqt-core not installed, installation required to access the CMake files."
        raise ImportError(msg) from None



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Evaluating the json file generated by the benchmarking script."""

from __future__ import annotations

import argparse
import json
import math
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypedDict

import pandas as pd

if TYPE_CHECKING:
    from os import PathLike

# Avoid output truncation
pd.set_option("display.max_colwidth", None)
pd.set_option("display.max_rows", None)
pd.set_option("display.width", None)

__sort_options = ["ratio", "algorithm"]
__higher_better_metrics = ["hits", "hit_ratio"]


class _BColors:
    """Class for colored output in the terminal."""

    OKGREEN = "\033[92m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"


def __flatten_dict(d: dict[Any, Any], parent_key: str = "", sep: str = ".") -> dict[str, Any]:
    """Flatten a nested dictionary. Every value only has one key which is the path to the value.

    Returns:
        A dictionary with the flattened keys and the values.
    """
    items = {}
    for key, value in d.items():
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(__flatten_dict(value, new_key, sep=sep))
        else:
            items[new_key] = value
    return items


class _BenchmarkDict(TypedDict, total=False):
    """A dictionary containing the data of one particular benchmark."""

    algo: str
    task: str
    n: int
    component: str
    metric: str


def __post_processing(key: str) -> _BenchmarkDict:
    """Postprocess the key of a flattened dictionary to get the metrics for the DataFrame columns.

    Returns:
        A dictionary containing the algorithm, task, number of qubits, component, and metric.

    Raises:
        ValueError: If the key is missing the algorithm, task, number of qubits, or metric.
    """
    metrics_divided = key.split(".")
    result_metrics = _BenchmarkDict()
    if len(metrics_divided) < 4:
        raise ValueError("Benchmark " + key + " is missing algorithm, task, number of qubits or metric!")
    result_metrics["algo"] = metrics_divided.pop(0)
    result_metrics["task"] = metrics_divided.pop(0)
    result_metrics["n"] = int(metrics_divided.pop(0))
    num_remaining_benchmarks = len(metrics_divided)
    if num_remaining_benchmarks == 1:
        result_metrics["component"] = ""
        result_metrics["metric"] = metrics_divided.pop(0)
    elif num_remaining_benchmarks == 2:
        if metrics_divided[0] == "dd":
            result_metrics["component"] = "" if metrics_divided[0] == "dd" else metrics_divided.pop(0)
            result_metrics["metric"] = metrics_divided[-1]
    else:
        separator = "_"
        # if the second-to-last element is not "total" then only the last element is the metric and the rest component
        if metrics_divided[-2] == "total":
            metric = separator.join(metrics_divided[-2:])
            result_metrics["metric"] = metric
            component = separator.join(metrics_divided[:-2])
            component = component.removeprefix("dd_")
            result_metrics["component"] = component
        else:
            result_metrics["metric"] = metrics_divided[-1]
            component = separator.join(metrics_divided[:-1])
            component = component.removeprefix("dd_")
            result_metrics["component"] = component

    return result_metrics


class _DataDict(TypedDict):
    """A dictionary containing the data for one entry in the DataFrame."""

    before: float
    after: float
    ratio: float
    algo: str
    task: str
    n: int
    component: str
    metric: str


def __aggregate(baseline_filepath: str | PathLike[str], feature_filepath: str | PathLike[str]) -> pd.DataFrame:
    """Aggregate the data from the baseline and feature json files into one DataFrame for visualization.

    Returns:
        A DataFrame containing the aggregated data.
    """
    base_path = Path(baseline_filepath)
    with base_path.open(mode="r", encoding="utf-8") as f:
        d = json.load(f)
    flattened_data = __flatten_dict(d)
    feature_path = Path(feature_filepath)
    with feature_path.open(mode="r", encoding="utf-8") as f:
        d_feature = json.load(f)
    flattened_feature = __flatten_dict(d_feature)

    for k, v in flattened_data.items():
        value = v
        if value == "unused":
            value = float("nan")
        if k in flattened_feature:
            ls = [value, flattened_feature[k]]
            flattened_data[k] = ls
            del flattened_feature[k]
        else:
            ls = [value, float("nan")]
            flattened_data[k] = ls
    # If a benchmark is in the feature file but not in the baseline file, it should be added with baseline marked as
    # "skipped"
    for k, v in flattened_feature.items():
        value = v
        if value == "unused":
            value = float("nan")
        ls = [float("nan"), value]
        flattened_data[k] = ls

    df_all_entries = []  # records individual entries

    for k, v in flattened_data.items():
        before = v[0]
        after = v[1]
        if math.isnan(before) or math.isnan(after):
            ratio = float("nan")
        else:
            ratio = after / before if before != 0 else 1 if after == 0 else math.inf
        key = k
        if k.endswith(tuple(__higher_better_metrics)):
            ratio = 1 / ratio if ratio != 0 else math.inf
            key += "*"
        before = round(before, 3) if isinstance(before, float) else before
        after = round(after, 3) if isinstance(after, float) else after
        ratio = round(ratio, 3)
        # postprocessing
        result_metrics = __post_processing(key)

        df_all_entries.append(
            _DataDict(
                before=before,
                after=after,
                ratio=ratio,
                algo=result_metrics["algo"],
                task=result_metrics["task"],
                n=result_metrics["n"],
                component=result_metrics["component"],
                metric=result_metrics["metric"],
            ),
        )

    df_all = pd.DataFrame(df_all_entries)
    df_all.index = pd.Index([""] * len(df_all.index))

    return df_all


def __print_results(
    *,
    df: pd.DataFrame,
    sort_indices: list[str],
    factor: float,
    no_split: bool,
    only_changed: bool,
) -> None:
    """Print the results in a nice table."""
    # after significantly smaller than before
    m1 = df["ratio"] < 1 - factor
    # after significantly larger than before
    m2 = df["ratio"] > 1 + factor
    # after is nan or before is nan or after is close to before
    m3 = (df["ratio"] != df["ratio"]) | ((1 - factor <= df["ratio"]) & (df["ratio"] <= 1 + factor))

    if no_split:
        if only_changed:
            print(df[m1 | m2].sort_values(by=sort_indices).to_markdown(index=False, stralign="right"))
        print(df.sort_values(by=sort_indices).to_markdown(index=False, stralign="right"))
        return

    print(f"\n{_BColors.OKGREEN}Benchmarks that have improved:{_BColors.ENDC}\n")
    print(df[m1].sort_values(by=sort_indices).to_markdown(index=False, stralign="right"))

    print(f"\n{_BColors.FAIL}Benchmarks that have worsened:{_BColors.ENDC}\n")
    print(df[m2].sort_values(by=sort_indices, ascending=False).to_markdown(index=False, stralign="right"))

    if only_changed:
        return

    print("\nBenchmarks that have stayed the same:\n")
    print(df[m3].sort_values(by=sort_indices).to_markdown(index=False, stralign="right"))


def compare(
    baseline_filepath: str | PathLike[str],
    feature_filepath: str | PathLike[str],
    *,
    factor: float = 0.1,
    sort: str = "ratio",
    dd: bool = False,
    only_changed: bool = False,
    no_split: bool = False,
    algorithm: str | None = None,
    task: str | None = None,
    num_qubits: int | None = None,
) -> None:
    """Compare the results of two benchmarking runs from the generated json file.

    Args:
        baseline_filepath: Path to the baseline json file.
        feature_filepath: Path to the feature json file.
        factor: How much a result has to change to be considered significant.
        sort: Sort the table by this column. Valid options are "ratio" and "algorithm".
        dd: Whether to show the detailed DD benchmark results.
        only_changed: Whether to only show results that changed significantly.
        no_split: Whether to merge all results together in one table or to separate the results
                  into benchmarks that improved, stayed the same, or worsened.
        algorithm: Only show results for this algorithm.
        task: Only show results for this task.
        num_qubits: Only show results for this number of qubits. Can only be used if algorithm is also specified.

    Raises:
        ValueError: If factor is negative or sort is invalid or if num_qubits is specified while algorithm is not.
    """
    if factor < 0:
        msg = "Factor must be positive!"
        raise ValueError(msg)
    if sort not in __sort_options:
        msg = "Invalid sort option! Valid options are 'ratio' and 'algorithm'."
        raise ValueError(msg)
    if algorithm is None and num_qubits is not None:
        msg = "num_qubits can only be specified if algorithm is also specified!"
        raise ValueError(msg)

    df_all = __aggregate(baseline_filepath, feature_filepath)

    if task is not None:
        df_all = df_all[df_all["task"].str.contains(task, case=False)]
    if algorithm is not None:
        df_all = df_all[df_all["algo"].str.contains(algorithm, case=False)]
    if num_qubits is not None:
        df_all = df_all[df_all["n"] == num_qubits]

    df_runtime = df_all[df_all["metric"] == "runtime"]
    df_runtime = df_runtime.drop(columns=["component", "metric"])
    print("\nRuntime:")
    sort_indices = ["ratio"] if sort == "ratio" else ["algo", "task", "n"]
    __print_results(
        df=df_runtime, sort_indices=sort_indices, factor=factor, no_split=no_split, only_changed=only_changed
    )

    if not dd:
        return

    print("\nDD Package details:")
    df_dd = df_all[df_all["metric"] != "runtime"]
    sort_indices = ["ratio"] if sort == "ratio" else ["algo", "task", "n", "component", "metric"]
    __print_results(df=df_dd, sort_indices=sort_indices, factor=factor, no_split=no_split, only_changed=only_changed)


def main() -> None:
    """Main function for the command line interface.

    This function is called when running the `mqt-core-compare` CLI command.

    .. code-block:: bash

        mqt-core-compare baseline.json feature.json [options]

    In addition to the mandatory filepath arguments, it provides the following optional command line options:

    - :code:`--factor <float>`: How much a result has to change to be considered significant.
    - :code:`--sort`: Sort the table by this column. Valid options are 'ratio' and 'algorithm'.
    - :code:`--dd`: Whether to show the detailed DD benchmark results.
    - :code:`--only_changed`: Whether to only show results that changed significantly.
    - :code:`--no_split`: Whether to merge all results together in one table or to separate the results into benchmarks
      that improved, stayed the same, or worsened.
    - :code:`--algorithm <str>`: Only show results for this algorithm.
    - :code:`--task <str>`: Only show results for this task.
    - :code:`--num_qubits <int>`: Only show results for this number of qubits.
      Can only be used if algorithm is also specified.
    """
    parser = argparse.ArgumentParser(
        description="Compare the results of two benchmarking runs from the generated json files.",
    )
    parser.add_argument("baseline_filepath", type=str, help="Path to the baseline json file.")
    parser.add_argument("feature_filepath", type=str, help="Path to the feature json file.")
    parser.add_argument(
        "--factor",
        type=float,
        default=0.1,
        help="How much a result has to change to be considered significant.",
    )
    parser.add_argument(
        "--sort",
        type=str,
        default="ratio",
        help="Sort the table by this column. Valid options are 'ratio' and 'algorithm'.",
    )
    parser.add_argument("--dd", action="store_true", help="Whether to show the detailed DD benchmark results.")
    parser.add_argument(
        "--only_changed",
        action="store_true",
        help="Whether to only show results that changed significantly.",
    )
    parser.add_argument(
        "--no_split",
        action="store_true",
        help="Whether to merge all results together in one table or to separate the results into "
        "benchmarks that improved, stayed the same, or worsened.",
    )
    parser.add_argument("--algorithm", type=str, help="Only show results for this algorithm.")
    parser.add_argument("--task", type=str, help="Only show results for this task.")
    parser.add_argument(
        "--num_qubits",
        type=int,
        help="Only show results for this number of qubits. Can only be used if algorithm is also specified.",
    )
    args = parser.parse_args()
    assert args is not None
    compare(
        args.baseline_filepath,
        args.feature_filepath,
        factor=args.factor,
        sort=args.sort,
        dd=args.dd,
        only_changed=args.only_changed,
        no_split=args.no_split,
        algorithm=args.algorithm,
        task=args.task,
        num_qubits=args.num_qubits,
    )



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""MQT Qiskit Plugin."""

from __future__ import annotations

from .mqt_to_qiskit import mqt_to_qiskit
from .qiskit_to_mqt import qiskit_to_mqt

__all__ = [
    "mqt_to_qiskit",
    "qiskit_to_mqt",
]



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Functionality for translating from the MQT to Qiskit."""

from __future__ import annotations

from typing import TYPE_CHECKING

from qiskit.circuit import (
    AncillaRegister,
    ClassicalRegister,
    IfElseOp,
    QuantumCircuit,
    QuantumRegister,
)
from qiskit.circuit.classical import expr
from qiskit.circuit.library import (
    DCXGate,
    ECRGate,
    HGate,
    IGate,
    PhaseGate,
    RGate,
    RXGate,
    RXXGate,
    RYGate,
    RYYGate,
    RZGate,
    RZXGate,
    RZZGate,
    SdgGate,
    SGate,
    SwapGate,
    SXdgGate,
    SXGate,
    TdgGate,
    TGate,
    U2Gate,
    U3Gate,
    XGate,
    XXMinusYYGate,
    XXPlusYYGate,
    YGate,
    ZGate,
    iSwapGate,
)
from qiskit.transpiler.layout import Layout, TranspileLayout

from ...ir import Permutation
from ...ir.operations import (
    ComparisonKind,
    CompoundOperation,
    Control,
    IfElseOperation,
    NonUnitaryOperation,
    OpType,
    StandardOperation,
)

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from qiskit.circuit import Clbit, Qubit
    from qiskit.circuit.singleton import SingletonGate

    from ...ir import QuantumComputation
    from ...ir.operations import Operation

__all__ = ["mqt_to_qiskit"]


def __dir__() -> list[str]:
    return __all__


def _translate_controls(controls: set[Control], qubit_map: Mapping[int, Qubit]) -> tuple[list[Qubit], str]:
    """Translate a set of :class:`~mqt.core.ir.operations.Control` to Qiskit.

    Args:
        controls: The controls to translate.
        qubit_map: A mapping from qubit indices to Qiskit :class:`~qiskit.circuit.Qubit`.

    Returns:
        A tuple containing the translated qubits and control states.
    """
    qubits: list[Qubit] = []
    ctrl_state: str = ""
    for control in controls:
        qubit = qubit_map[control.qubit]
        qubits.append(qubit)
        # MSB to the left
        ctrl_state = "1" + ctrl_state if control.type_ == Control.Type.Pos else "0" + ctrl_state
    return qubits, ctrl_state


def _translate_targets(targets: Sequence[int], qubit_map: Mapping[int, Qubit]) -> list[Qubit]:
    """Translate a sequence of target qubit indices to a list of Qiskit :class:`~qiskit.circuit.Qubit`.

    Args:
        targets: The target qubit indices to translate.
        qubit_map: A mapping from qubit indices to Qiskit :class:`~qiskit.circuit.Qubit`.

    Returns:
        The translated qubits.
    """
    return [qubit_map[target] for target in targets]


def _add_standard_operation(circ: QuantumCircuit, op: StandardOperation, qubit_map: Mapping[int, Qubit]) -> None:
    """Add a :class:`~mqt.core.ir.operations.StandardOperation`.

    Args:
        circ: The Qiskit circuit to add the operation to.
        op: The MQT operation to add.
        qubit_map: A mapping from qubit indices to Qiskit :class:`~qiskit.circuit.Qubit`.

    Raises:
        TypeError: If the operation type is not supported.
    """
    targets = _translate_targets(op.targets, qubit_map)

    if op.type_ == OpType.barrier:
        circ.barrier(targets)
        return

    controls, ctrl_state = _translate_controls(op.controls, qubit_map)

    gate_map_singleton: dict[OpType, SingletonGate] = {
        OpType.i: IGate(),
        OpType.x: XGate(),
        OpType.y: YGate(),
        OpType.z: ZGate(),
        OpType.h: HGate(),
        OpType.s: SGate(),
        OpType.sdg: SdgGate(),
        OpType.t: TGate(),
        OpType.tdg: TdgGate(),
        OpType.sx: SXGate(),
        OpType.sxdg: SXdgGate(),
        OpType.dcx: DCXGate(),
        OpType.ecr: ECRGate(),
        OpType.swap: SwapGate(),
        OpType.iswap: iSwapGate(),
    }

    if op.type_ in gate_map_singleton:
        gate = gate_map_singleton[op.type_]
        if len(controls) == 0:
            circ.append(gate, targets)
        else:
            circ.append(gate.control(len(controls), ctrl_state=ctrl_state), [*controls, *targets])
        return

    gate_map_single_param: dict[OpType, type] = {
        OpType.rx: RXGate,
        OpType.ry: RYGate,
        OpType.rz: RZGate,
        OpType.p: PhaseGate,
        OpType.rxx: RXXGate,
        OpType.ryy: RYYGate,
        OpType.rzz: RZZGate,
        OpType.rzx: RZXGate,
    }

    if op.type_ in gate_map_single_param:
        gate = gate_map_single_param[op.type_]
        parameter = op.parameter[0]
        if len(controls) == 0:
            circ.append(gate(parameter), targets)
        else:
            circ.append(gate(parameter).control(len(controls), ctrl_state=ctrl_state), [*controls, *targets])
        return

    gate_map_two_param: dict[OpType, type] = {
        OpType.u2: U2Gate,
        OpType.r: RGate,
        OpType.xx_plus_yy: XXPlusYYGate,
        OpType.xx_minus_yy: XXMinusYYGate,
    }

    if op.type_ in gate_map_two_param:
        gate = gate_map_two_param[op.type_]
        parameter1, parameter2 = op.parameter
        if len(controls) == 0:
            circ.append(gate(parameter1, parameter2), targets)
        else:
            circ.append(
                gate(parameter1, parameter2).control(len(controls), ctrl_state=ctrl_state), [*controls, *targets]
            )
        return

    gate_map_three_param: dict[OpType, type] = {
        OpType.u: U3Gate,
    }

    if op.type_ in gate_map_three_param:
        gate = gate_map_three_param[op.type_]
        parameter1, parameter2, parameter3 = op.parameter
        if len(controls) == 0:
            circ.append(gate(parameter1, parameter2, parameter3), targets)
        else:
            circ.append(
                gate(parameter1, parameter2, parameter3).control(len(controls), ctrl_state=ctrl_state),
                [*controls, *targets],
            )
        return

    msg = f"Unsupported operation type: {op.type_}"
    raise TypeError(msg)


def _add_non_unitary_operation(
    circ: QuantumCircuit,
    op: NonUnitaryOperation,
    qubit_map: Mapping[int, Qubit],
    clbit_map: Mapping[int, Clbit],
) -> None:
    """Add a :class:`~mqt.core.ir.operations.NonUnitaryOperation`.

    Args:
        circ: The Qiskit circuit to add the operation to.
        op: The MQT operation to add.
        qubit_map: A mapping from qubit indices to Qiskit :class:`~qiskit.circuit.Qubit`.
        clbit_map: A mapping from classical bit indices to Qiskit :class:`~qiskit.circuit.Clbit`.
    """
    if op.type_ == OpType.measure:
        for qubit, clbit in zip(op.targets, op.classics, strict=False):
            circ.measure(qubit_map[qubit], clbit_map[clbit])
        return

    if op.type_ == OpType.reset:
        for qubit in op.targets:
            circ.reset(qubit_map[qubit])
        return


def _add_compound_operation(
    circ: QuantumCircuit,
    op: CompoundOperation,
    qubit_map: Mapping[int, Qubit],
    clbit_map: Mapping[int, Clbit],
) -> None:
    """Add a :class:`~mqt.core.ir.operations.CompoundOperation`.

    Args:
        circ: The Qiskit circuit to add the operation to.
        op: The MQT operation to add.
        qubit_map: A mapping from qubit indices to Qiskit :class:`~qiskit.circuit.Qubit`.
        clbit_map: A mapping from classical bit indices to Qiskit :class:`~qiskit.circuit.Clbit`.
    """
    inner_circ = QuantumCircuit(*circ.qregs, *circ.cregs)
    for inner_op in op:
        _add_operation(inner_circ, inner_op, qubit_map, clbit_map)
    circ.append(inner_circ.to_instruction(), circ.qubits, circ.clbits)


def _add_if_else_operation(
    circ: QuantumCircuit,
    op: IfElseOperation,
    qubit_map: Mapping[int, Qubit],
    clbit_map: Mapping[int, Clbit],
) -> None:
    """Add a :class:`~.IfElseOperation`.

    Args:
        circ: The Qiskit circuit to add the operation to.
        op: The :class:`~.IfElseOperation` operation to add.
        qubit_map: A mapping from qubit indices to Qiskit :class:`~qiskit.circuit.Qubit`.
        clbit_map: A mapping from classical bit indices to Qiskit :class:`~qiskit.circuit.Clbit`.
    """
    if op.control_register is not None:
        left_hand_side = next(reg for reg in circ.cregs if reg.name == op.control_register.name)
        right_hand_side = op.expected_value_register
    else:
        assert op.control_bit is not None
        left_hand_side = clbit_map[op.control_bit]
        right_hand_side = op.expected_value_bit

    condition: expr.Expr | tuple[ClassicalRegister | Clbit, int]
    if op.comparison_kind == ComparisonKind.eq:
        # directly handle equality conditions without expressions
        condition = (left_hand_side, right_hand_side)
    elif op.comparison_kind == ComparisonKind.neq:
        condition = expr.not_equal(left_hand_side, right_hand_side)
    elif op.comparison_kind == ComparisonKind.lt:
        condition = expr.less(left_hand_side, right_hand_side)
    elif op.comparison_kind == ComparisonKind.leq:
        condition = expr.less_equal(left_hand_side, right_hand_side)
    elif op.comparison_kind == ComparisonKind.gt:
        condition = expr.greater(left_hand_side, right_hand_side)
    elif op.comparison_kind == ComparisonKind.geq:
        condition = expr.greater_equal(left_hand_side, right_hand_side)

    then_circ = QuantumCircuit(*circ.qregs, *circ.cregs)
    _add_operation(then_circ, op.then_operation, qubit_map, clbit_map)

    else_circ: QuantumCircuit | None = None
    if op.else_operation is not None:
        else_circ = QuantumCircuit(*circ.qregs, *circ.cregs)
        _add_operation(else_circ, op.else_operation, qubit_map, clbit_map)

    circ.append(IfElseOp(condition, then_circ, else_circ), circ.qubits, circ.clbits)


def _add_operation(
    circ: QuantumCircuit,
    op: Operation,
    qubit_map: Mapping[int, Qubit],
    clbit_map: Mapping[int, Clbit],
) -> None:
    """Add an operation to a Qiskit circuit.

    Args:
        circ: The Qiskit circuit to add the operation to.
        op: The MQT operation to add.
        qubit_map: A mapping from qubit indices to Qiskit :class:`~qiskit.circuit.Qubit`.
        clbit_map: A mapping from classical bit indices to Qiskit :class:`~qiskit.circuit.Clbit`.

    Raises:
        TypeError: If the operation type is not supported.
    """
    if isinstance(op, StandardOperation):
        _add_standard_operation(circ, op, qubit_map)
    elif isinstance(op, NonUnitaryOperation):
        _add_non_unitary_operation(circ, op, qubit_map, clbit_map)
    elif isinstance(op, CompoundOperation):
        _add_compound_operation(circ, op, qubit_map, clbit_map)
    elif isinstance(op, IfElseOperation):
        _add_if_else_operation(circ, op, qubit_map, clbit_map)
    else:
        msg = f"Unsupported operation type: {type(op)}"
        raise TypeError(msg)


def mqt_to_qiskit(qc: QuantumComputation, *, set_layout: bool = False) -> QuantumCircuit:
    """Convert a :class:`~mqt.core.ir.QuantumComputation` to a Qiskit :class:`~qiskit.circuit.QuantumCircuit`.

    Args:
        qc: The MQT circuit to convert.
        set_layout: If true, the :attr:`~qiskit.circuit.QuantumCircuit.layout` property is populated with the
                    initial layout and output permutation of the MQT circuit.

    Returns:
        The converted circuit.

    Raises:
        NotImplementedError: If the MQT circuit contains variables.
    """
    if not qc.is_variable_free():
        msg = "Converting symbolic circuits with variables to Qiskit is not yet supported."
        raise NotImplementedError(msg)

    circ = QuantumCircuit()

    if qc.name is not None:
        circ.name = qc.name

    qregs = sorted((qc.qregs | qc.ancregs).values(), key=lambda reg: reg.start)
    qubit_map: dict[int, Qubit] = {}
    for qreg in qregs:
        qiskit_reg = (
            QuantumRegister(size=qreg.size, name=qreg.name)
            if qreg.name in qc.qregs
            else AncillaRegister(size=qreg.size, name=qreg.name)
        )
        circ.add_register(qiskit_reg)
        for i, qubit in enumerate(qiskit_reg):
            qubit_map[qreg.start + i] = qubit

    cregs = sorted(qc.cregs.values(), key=lambda reg: reg.start)
    clbit_map: dict[int, Clbit] = {}
    for creg in cregs:
        qiskit_creg = ClassicalRegister(size=creg.size, name=creg.name)
        circ.add_register(qiskit_creg)
        for i, clbit in enumerate(qiskit_creg):
            clbit_map[creg.start + i] = clbit

    for op in qc:
        _add_operation(circ, op, qubit_map, clbit_map)

    if not set_layout:
        return circ

    # create a list of physical qubits initialized to none, but with the correct length
    p2v: list[Qubit | None] = [None] * len(circ.qubits)
    # fill the list with the correct virtual qubits
    for virtual, physical in qc.initial_layout.items():
        p2v[virtual] = qubit_map[physical]
    initial_layout = Layout().from_qubit_list(p2v, *circ.qregs)

    # reconstruct the final layout, which is the permutation between the initial layout and the output permutation
    permutation = Permutation()
    for physical, virtual in qc.output_permutation.items():
        # find the virtual qubit in the initial layout and store the corresponding physical qubit
        for p, v in qc.initial_layout.items():
            if v == virtual:
                permutation[p] = physical
                continue

    p2v = [None] * len(circ.qubits)
    # fill the list with the correct virtual qubits
    for physical, virtual in permutation.items():
        p2v[virtual] = qubit_map[physical]
    final_layout = Layout().from_qubit_list(p2v, *circ.qregs)

    circ._layout = TranspileLayout(  # noqa: SLF001
        initial_layout=initial_layout,
        input_qubit_mapping={qubit: idx for idx, qubit in qubit_map.items()},
        final_layout=final_layout,
        _input_qubit_count=qc.num_qubits,
        _output_qubit_list=list(final_layout.get_virtual_bits()),
    )

    return circ



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if sys.version_info >= (3, 11):
    from typing import assert_never
elif TYPE_CHECKING:
    from typing_extensions import assert_never
else:

    def assert_never(_: object) -> None:
        msg = "Expected code to be unreachable"
        raise AssertionError(msg)


__all__ = ["assert_never"]


def __dir__() -> list[str]:
    return __all__



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

from collections.abc import Iterable
from enum import Enum

__all__ = [
    "Device",
    "devices",
]

class Device:
    """A device represents a quantum device with its properties and capabilities."""
    class Status(Enum):
        offline = ...
        idle = ...
        busy = ...
        error = ...
        maintenance = ...
        calibration = ...

    class Site:
        """A site represents a potential qubit location on a quantum device."""
        def index(self) -> int:
            """Returns the index of the site."""
        def t1(self) -> int | None:
            """Returns the T1 coherence time of the site."""
        def t2(self) -> int | None:
            """Returns the T2 coherence time of the site."""
        def name(self) -> str | None:
            """Returns the name of the site."""
        def x_coordinate(self) -> int | None:
            """Returns the x coordinate of the site."""
        def y_coordinate(self) -> int | None:
            """Returns the y coordinate of the site."""
        def z_coordinate(self) -> int | None:
            """Returns the z coordinate of the site."""
        def is_zone(self) -> bool | None:
            """Returns whether the site is a zone."""
        def x_extent(self) -> int | None:
            """Returns the x extent of the site."""
        def y_extent(self) -> int | None:
            """Returns the y extent of the site."""
        def z_extent(self) -> int | None:
            """Returns the z extent of the site."""
        def module_index(self) -> int | None:
            """Returns the index of the module the site belongs to."""
        def submodule_index(self) -> int | None:
            """Returns the index of the submodule the site belongs to."""
        def __eq__(self, other: object) -> bool:
            """Checks if two sites are equal."""
        def __ne__(self, other: object) -> bool:
            """Checks if two sites are not equal."""

    class Operation:
        """An operation represents a quantum operation that can be performed on a quantum device."""
        def name(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> str:
            """Returns the name of the operation."""
        def qubits_num(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> int | None:
            """Returns the number of qubits the operation acts on."""
        def parameters_num(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> int:
            """Returns the number of parameters the operation has."""
        def duration(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> int | None:
            """Returns the duration of the operation."""
        def fidelity(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> float | None:
            """Returns the fidelity of the operation."""
        def interaction_radius(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> int | None:
            """Returns the interaction radius of the operation."""
        def blocking_radius(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> int | None:
            """Returns the blocking radius of the operation."""
        def idling_fidelity(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> float | None:
            """Returns the idling fidelity of the operation."""
        def is_zoned(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> bool | None:
            """Returns whether the operation is zoned."""
        def sites(
            self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...
        ) -> Iterable[Device.Site] | None:
            """Returns the list of sites the operation can be performed on."""
        def mean_shuttling_speed(self, sites: Iterable[Device.Site] = ..., params: Iterable[float] = ...) -> int | None:
            """Returns the mean shuttling speed of the operation."""
        def __eq__(self, other: object) -> bool:
            """Checks if two operations are equal."""
        def __ne__(self, other: object) -> bool:
            """Checks if two operations are not equal."""

    def name(self) -> str:
        """Returns the name of the device."""
    def version(self) -> str:
        """Returns the version of the device."""
    def status(self) -> Device.Status:
        """Returns the current status of the device."""
    def library_version(self) -> str:
        """Returns the version of the library used to define the device."""
    def qubits_num(self) -> int:
        """Returns the number of qubits available on the device."""
    def sites(self) -> Iterable[Site]:
        """Returns the list of sites available on the device."""
    def operations(self) -> Iterable[Operation]:
        """Returns the list of operations supported by the device."""
    def coupling_map(self) -> Iterable[tuple[Site, Site]] | None:
        """Returns the coupling map of the device as a list of site pairs."""
    def needs_calibration(self) -> int | None:
        """Returns whether the device needs calibration."""
    def length_unit(self) -> str | None:
        """Returns the unit of length used by the device."""
    def length_scale_factor(self) -> float | None:
        """Returns the scale factor for length used by the device."""
    def duration_unit(self) -> str | None:
        """Returns the unit of duration used by the device."""
    def duration_scale_factor(self) -> float | None:
        """Returns the scale factor for duration used by the device."""
    def min_atom_distance(self) -> int | None:
        """Returns the minimum atom distance on the device."""
    def __eq__(self, other: object) -> bool:
        """Checks if two devices are equal."""
    def __ne__(self, other: object) -> bool:
        """Checks if two devices are not equal."""

def devices() -> Iterable[Device]:
    """Returns a list of available devices."""



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Command line interface for mqt-core."""

from __future__ import annotations

import argparse
import sys

from ._commands import cmake_dir, include_dir
from ._version import version as __version__


def main() -> None:
    """Entry point for the mqt-core command line interface.

    This function is called when running the `mqt-core-cli` script.

    .. code-block:: bash

        mqt-core-cli [--version] [--include_dir] [--cmake_dir]

    It provides the following command line options:

    - :code:`--version`: Print the version and exit.
    - :code:`--include_dir`: Print the path to the mqt-core C++ include directory.
    - :code:`--cmake_dir`: Print the path to the mqt-core CMake module directory.

    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=__version__, help="Print version and exit.")

    parser.add_argument(
        "--include_dir", action="store_true", help="Print the path to the mqt-core C++ include directory."
    )
    parser.add_argument(
        "--cmake_dir", action="store_true", help="Print the path to the mqt-core CMake module directory."
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.include_dir:
        print(include_dir().resolve())
    if args.cmake_dir:
        print(cmake_dir().resolve())


if __name__ == "__main__":
    main()



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

"""Functionality for translating from Qiskit to the MQT."""

from __future__ import annotations

import re
import warnings
from typing import TYPE_CHECKING, cast

from qiskit.circuit import AncillaRegister, Clbit
from qiskit.circuit import ClassicalRegister as QiskitClassicalRegister
from qiskit.circuit.classical import expr

from ...ir import QuantumComputation
from ...ir.operations import (
    ComparisonKind,
    CompoundOperation,
    Control,
    IfElseOperation,
    NonUnitaryOperation,
    OpType,
    StandardOperation,
    SymbolicOperation,
)
from ...ir.symbolic import Expression, Term, Variable

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from qiskit.circuit import IfElseOp, Instruction, ParameterExpression, QuantumCircuit, Qubit

    from ...ir.operations import Operation
    from ...ir.registers import ClassicalRegister


__all__ = ["qiskit_to_mqt"]


def __dir__() -> list[str]:
    return __all__


def qiskit_to_mqt(circ: QuantumCircuit) -> QuantumComputation:
    """Convert a Qiskit :class:`~qiskit.circuit.QuantumCircuit` to a :class:`~mqt.core.ir.QuantumComputation` object.

    Args:
        circ: The Qiskit circuit to convert.

    Returns:
        The converted circuit.
    """
    qc = QuantumComputation()

    if circ.name is not None:
        qc.name = circ.name

    qubit_index = 0
    qubit_map: dict[Qubit, int] = {}
    for reg in circ.qregs:
        size = reg.size
        if size == 0:
            continue
        if isinstance(reg, AncillaRegister):
            qc.add_ancillary_register(size, reg.name)
        else:
            qc.add_qubit_register(size, reg.name)
        for qubit in reg:
            qubit_map[qubit] = qubit_index
            qubit_index += 1

    clbit_index = 0
    clbit_map: dict[Clbit, int] = {}
    for reg in circ.cregs:
        size = reg.size
        if size == 0:
            continue
        qc.add_classical_register(size, reg.name)
        for bit in reg:
            clbit_map[bit] = clbit_index
            clbit_index += 1

    try:
        qc.global_phase = circ.global_phase
    except TypeError:
        warnings.warn(
            "Symbolic global phase values are not supported yet. Setting global phase to 0.",
            RuntimeWarning,
            stacklevel=2,
        )
        qc.global_phase = 0

    for instruction in circ.data:
        symb_params = _emplace_operation(
            qc,
            instruction.operation,
            instruction.qubits,
            instruction.clbits,
            instruction.operation.params,
            qubit_map,
            clbit_map,
            qc.cregs,
        )
        for symb_param in symb_params:
            qc.add_variable(symb_param)

    # import initial layout and output permutation if available
    if circ.layout is not None:
        _import_layouts(qc, circ)

    qc.initialize_io_mapping()
    return qc


_NATIVELY_SUPPORTED_GATES = frozenset({
    "i",
    "id",
    "iden",
    "x",
    "y",
    "z",
    "h",
    "s",
    "sdg",
    "t",
    "tdg",
    "p",
    "u1",
    "r",
    "rx",
    "ry",
    "rz",
    "u2",
    "u",
    "u3",
    "cx",
    "cy",
    "cz",
    "cs",
    "csdg",
    "cp",
    "cu1",
    "ch",
    "crx",
    "cry",
    "crz",
    "cu3",
    "ccx",
    "swap",
    "cswap",
    "iswap",
    "sx",
    "sxdg",
    "csx",
    "mcx",
    # we keep the "mcx_*" variants here for compatibility with Qiskit 1.x
    "mcx_gray",
    "mcx_recursive",
    "mcx_vchain",
    "mcphase",
    "mcrx",
    "mcry",
    "mcrz",
    "dcx",
    "ecr",
    "rxx",
    "ryy",
    "rzx",
    "rzz",
    "xx_minus_yy",
    "xx_plus_yy",
    "if_else",
    "reset",
    "barrier",
    "measure",
})


def _emplace_operation(
    qc: QuantumComputation | CompoundOperation,
    instr: Instruction,
    qargs: Sequence[Qubit],
    cargs: Sequence[Clbit],
    params: Sequence[float | ParameterExpression],
    qubit_map: Mapping[Qubit, int],
    clbit_map: Mapping[Clbit, int],
    cregs: Mapping[str, ClassicalRegister],
) -> list[float | ParameterExpression]:
    name = instr.name

    if name not in _NATIVELY_SUPPORTED_GATES:
        try:
            return _import_definition(qc, instr.definition, qargs, cargs, qubit_map, clbit_map, cregs)
        except Exception as ex:  # PRAGMA: NO COVER
            msg = f"Unsupported gate {name} with definition {instr.definition}"
            raise NotImplementedError(msg) from ex

    qubits = [qubit_map[qubit] for qubit in qargs]

    if name == "measure":
        clbits = [clbit_map[clbit] for clbit in cargs]
        qc.append(NonUnitaryOperation(qubits, clbits))
        return []

    if name == "reset":
        qc.append(NonUnitaryOperation(qubits))
        return []

    if name == "barrier":
        qc.append(StandardOperation(qubits, OpType.barrier))
        return []

    if name in {"i", "id", "iden"}:
        return _add_operation(qc, OpType.i, qargs, params, qubit_map)

    # we keep "mcx_gray" here for compatibility reasons with older Qiskit versions, e.g., 1.0.0 where "noancilla"
    # is the default for the argument "mode" which leads to the gate name "gray_code"
    if name in {"x", "cx", "ccx", "mcx", "mcx_gray"}:
        return _add_operation(qc, OpType.x, qargs, params, qubit_map)

    if name in {"y", "cy"}:
        return _add_operation(qc, OpType.y, qargs, params, qubit_map)

    if name in {"z", "cz"}:
        return _add_operation(qc, OpType.z, qargs, params, qubit_map)

    if name in {"h", "ch"}:
        return _add_operation(qc, OpType.h, qargs, params, qubit_map)

    if name in {"s", "cs"}:
        return _add_operation(qc, OpType.s, qargs, params, qubit_map)

    if name in {"sdg", "csdg"}:
        return _add_operation(qc, OpType.sdg, qargs, params, qubit_map)

    if name == "t":
        return _add_operation(qc, OpType.t, qargs, params, qubit_map)

    if name == "tdg":
        return _add_operation(qc, OpType.tdg, qargs, params, qubit_map)

    if name in {"sx", "csx"}:
        return _add_operation(qc, OpType.sx, qargs, params, qubit_map)

    if name == "sxdg":
        return _add_operation(qc, OpType.sxdg, qargs, params, qubit_map)

    if name == "mcx_recursive":
        if len(qargs) <= 5:
            return _add_operation(qc, OpType.x, qargs, params, qubit_map)
        # reconfigure controls and targets (drops the last qubit as ancilla)
        qargs = qargs[:-1]
        return _add_operation(qc, OpType.x, qargs, params, qubit_map)

    if name == "mcx_vchain":
        size = len(qargs)
        num_controls = (size + 1) // 2
        # reconfigure controls and targets (drops the last num_controls - 2 qubits as ancilla)
        if num_controls > 2:
            qargs = qargs[: -num_controls + 2]
        return _add_operation(qc, OpType.x, qargs, params, qubit_map)

    if name in {"r", "prx"}:
        return _add_operation(qc, OpType.r, qargs, params, qubit_map)

    if name in {"rx", "crx", "mcrx"}:
        return _add_operation(qc, OpType.rx, qargs, params, qubit_map)

    if name in {"ry", "cry", "mcry"}:
        return _add_operation(qc, OpType.ry, qargs, params, qubit_map)

    if name in {"rz", "crz", "mcrz"}:
        return _add_operation(qc, OpType.rz, qargs, params, qubit_map)

    if name in {"p", "u1", "cp", "cu1", "mcphase"}:
        return _add_operation(qc, OpType.p, qargs, params, qubit_map)

    if name == "u2":
        return _add_operation(qc, OpType.u2, qargs, params, qubit_map)

    if name in {"u", "u3", "cu3"}:
        return _add_operation(qc, OpType.u, qargs, params, qubit_map)

    if name in {"swap", "cswap"}:
        return _add_two_target_operation(qc, OpType.swap, qargs, params, qubit_map)

    if name == "iswap":
        return _add_two_target_operation(qc, OpType.iswap, qargs, params, qubit_map)

    if name == "dcx":
        return _add_two_target_operation(qc, OpType.dcx, qargs, params, qubit_map)

    if name == "ecr":
        return _add_two_target_operation(qc, OpType.ecr, qargs, params, qubit_map)

    if name == "rxx":
        return _add_two_target_operation(qc, OpType.rxx, qargs, params, qubit_map)

    if name == "ryy":
        return _add_two_target_operation(qc, OpType.ryy, qargs, params, qubit_map)

    if name == "rzz":
        return _add_two_target_operation(qc, OpType.rzz, qargs, params, qubit_map)

    if name == "rzx":
        return _add_two_target_operation(qc, OpType.rzx, qargs, params, qubit_map)

    if name == "xx_minus_yy":
        return _add_two_target_operation(qc, OpType.xx_minus_yy, qargs, params, qubit_map)

    if name == "xx_plus_yy":
        return _add_two_target_operation(qc, OpType.xx_plus_yy, qargs, params, qubit_map)

    if name == "if_else":
        return _add_if_else_operation(qc, instr, qargs, cargs, qubit_map, clbit_map, cregs)

    msg = f"Unsupported gate {name}"  # pragma: no cover
    raise NotImplementedError(msg)


_SUM_REGEX = re.compile(r"[+|-]?[^+-]+")
_PROD_REGEX = re.compile(r"[*/]?[^*/]+")


def _parse_symbolic_expression(qiskit_expr: ParameterExpression | float) -> float | Expression:
    if isinstance(qiskit_expr, float):
        return qiskit_expr

    expr_str = str(qiskit_expr).strip()
    expr = Expression()
    is_const = True
    for summand in _SUM_REGEX.findall(expr_str):
        sign = 1
        summand_no_operator = summand
        if summand[0] == "+":
            summand_no_operator = summand[1:]
        elif summand[0] == "-":
            summand_no_operator = summand[1:]
            sign = -1

        coeff = 1.0
        var = ""
        for factor in _PROD_REGEX.findall(summand_no_operator):
            is_div = False
            factor_no_operator = factor
            if factor[0] == "*":
                factor_no_operator = factor[1:]
            elif factor[0] == "/":
                factor_no_operator = factor[1:]
                is_div = True

            factor_no_operator = factor_no_operator.strip()
            if factor_no_operator.replace(".", "").isnumeric():
                f = float(factor_no_operator)
                coeff *= 1.0 / f if is_div else f
            else:
                var = factor_no_operator

        if not var:
            expr += coeff
        else:
            is_const = False
            expr += Term(Variable(var), sign * coeff)

    if is_const:
        return expr.constant
    return expr


def _add_operation(
    qc: QuantumComputation | CompoundOperation,
    type_: OpType,
    qargs: Sequence[Qubit],
    params: Sequence[float | ParameterExpression],
    qubit_map: Mapping[Qubit, int],
) -> list[float | ParameterExpression]:
    qubits = [qubit_map[qubit] for qubit in qargs]
    target = qubits.pop()
    controls = {Control(qubit) for qubit in qubits}
    parameters = [_parse_symbolic_expression(param) for param in params]
    if any(isinstance(parameter, Expression) for parameter in parameters):
        qc.append(SymbolicOperation(controls, target, type_, parameters))
    else:
        qc.append(StandardOperation(controls, target, type_, cast("list[float]", parameters)))
    return parameters


def _add_two_target_operation(
    qc: QuantumComputation | CompoundOperation,
    type_: OpType,
    qargs: Sequence[Qubit],
    params: Sequence[float | ParameterExpression],
    qubit_map: Mapping[Qubit, int],
) -> list[float | ParameterExpression]:
    qubits = [qubit_map[qubit] for qubit in qargs]
    target2 = qubits.pop()
    target1 = qubits.pop()
    controls = {Control(qubit) for qubit in qubits}
    parameters = [_parse_symbolic_expression(param) for param in params]
    if any(isinstance(parameter, Expression) for parameter in parameters):
        qc.append(SymbolicOperation(controls, target1, target2, type_, parameters))
    else:
        qc.append(StandardOperation(controls, target1, target2, type_, cast("list[float]", parameters)))
    return parameters


def _add_if_else_operation(
    qc: QuantumComputation | CompoundOperation,
    if_else_op: IfElseOp,
    qargs: Sequence[Qubit],
    cargs: Sequence[Clbit],
    qubit_map: Mapping[Qubit, int],
    clbit_map: Mapping[Clbit, int],
    cregs: Mapping[str, ClassicalRegister],
) -> list[float | ParameterExpression]:
    then_operation_compound = CompoundOperation()
    then_params = _import_definition(
        then_operation_compound,
        if_else_op.params[0],
        qargs,
        cargs,
        qubit_map,
        clbit_map,
        cregs,
    )
    then_operation: Operation = then_operation_compound[0]
    if isinstance(then_operation, CompoundOperation) and len(then_operation) == 1:
        then_operation = then_operation[0]

    else_operation: Operation | None = None
    else_params: list[float | ParameterExpression] = []
    if if_else_op.params[1] is not None:
        else_operation_compound = CompoundOperation()
        else_params = _import_definition(
            else_operation_compound,
            if_else_op.params[1],
            qargs,
            cargs,
            qubit_map,
            clbit_map,
            cregs,
        )
        else_operation = else_operation_compound[0]
        if isinstance(else_operation, CompoundOperation) and len(else_operation) == 1:
            else_operation = else_operation[0]

    condition = if_else_op.condition
    if isinstance(condition, tuple):
        assert len(condition) == 2
        classical_control = condition[0]
        expected_value = condition[1]
        comparison_kind = ComparisonKind.eq
    else:
        assert isinstance(condition, expr.Binary)
        classical_control = condition.left.var
        expected_value = condition.right.value
        if condition.op == expr.Binary.Op.EQUAL:
            comparison_kind = ComparisonKind.eq
        elif condition.op == expr.Binary.Op.NOT_EQUAL:
            comparison_kind = ComparisonKind.neq
        elif condition.op == expr.Binary.Op.LESS:
            comparison_kind = ComparisonKind.lt
        elif condition.op == expr.Binary.Op.LESS_EQUAL:
            comparison_kind = ComparisonKind.leq
        elif condition.op == expr.Binary.Op.GREATER:
            comparison_kind = ComparisonKind.gt
        elif condition.op == expr.Binary.Op.GREATER_EQUAL:
            comparison_kind = ComparisonKind.geq
        else:
            msg = f"Unsupported binary operation {condition.op}"
            raise TypeError(msg)

    if isinstance(classical_control, Clbit):
        if_else_operation = IfElseOperation(
            then_operation,
            else_operation,
            clbit_map[classical_control],
            bool(expected_value),
            comparison_kind,
        )
    else:
        assert isinstance(classical_control, QiskitClassicalRegister)
        if_else_operation = IfElseOperation(
            then_operation,
            else_operation,
            cregs[classical_control.name],
            int(expected_value),
            comparison_kind,
        )
    qc.append(if_else_operation)

    return then_params + else_params


def _import_layouts(qc: QuantumComputation, circ: QuantumCircuit) -> None:
    qc.initial_layout.clear()
    initial_index_layout = circ.layout.initial_index_layout()
    for virtual, physical in enumerate(initial_index_layout):
        qc.initial_layout[physical] = virtual

    # Properly mark ancillary qubits
    for register in circ.layout.initial_layout.get_registers():
        if register.name != "ancilla" and not isinstance(register, AncillaRegister):
            continue
        for qubit in reversed(register):
            physical_qubit = circ.layout.initial_layout[qubit]
            virtual_qubit = qc.initial_layout[physical_qubit]
            qc.set_circuit_qubit_ancillary(virtual_qubit)

    if circ.layout.final_layout is None:
        qc.output_permutation = qc.initial_layout
        return

    qc.output_permutation.clear()
    final_index_layout = circ.layout.final_index_layout()
    for virtual, physical in enumerate(final_index_layout):
        qc.output_permutation[physical] = virtual

    # Properly mark garbage qubits
    # Any qubit in the initial layout that is not in the final layout is garbage
    for virtual_qubit in range(len(final_index_layout), len(initial_index_layout)):
        qc.set_circuit_qubit_garbage(virtual_qubit)


def _import_definition(
    qc: QuantumComputation | CompoundOperation,
    circ: QuantumCircuit,
    qargs: Sequence[Qubit],
    cargs: Sequence[Clbit],
    qubit_map: Mapping[Qubit, int],
    clbit_map: Mapping[Clbit, int],
    cregs: Mapping[str, ClassicalRegister],
) -> list[float | ParameterExpression]:
    qarg_map = dict(zip(circ.qubits, qargs, strict=False))
    carg_map = dict(zip(circ.clbits, cargs, strict=False))

    qc.append(CompoundOperation())
    comp_op = cast("CompoundOperation", qc[-1])

    params = []
    for instruction in circ.data:
        mapped_qargs = [qarg_map[qarg] for qarg in instruction.qubits]
        mapped_cargs = [carg_map[carg] for carg in instruction.clbits]
        operation = instruction.operation
        new_params = _emplace_operation(
            comp_op,
            operation,
            mapped_qargs,
            mapped_cargs,
            operation.params,
            qubit_map,
            clbit_map,
            cregs,
        )
        params.extend(new_params)
    return params



# Copyright (c) 2023 - 2025 Chair for Design Automation, TUM
# Copyright (c) 2025 Munich Quantum Software Company GmbH
# All rights reserved.
#
# SPDX-License-Identifier: MIT
#
# Licensed under the MIT License

from collections.abc import Iterator, Mapping, Sequence
from typing import overload

__all__ = ["Expression", "Term", "Variable"]

class Variable:
    """A symbolic variable.

    Args:
        name: The name of the variable.

    Note:
        Variables are uniquely identified by their name, so if a variable with the same name already exists,
        the existing variable will be returned.
    """

    def __eq__(self, arg0: object) -> bool: ...
    def __gt__(self, arg0: Variable) -> bool: ...
    def __hash__(self) -> int: ...
    def __init__(self, name: str = "") -> None: ...
    def __lt__(self, arg0: Variable) -> bool: ...
    def __ne__(self, arg0: object) -> bool: ...
    @property
    def name(self) -> str:
        """The name of the variable."""

class Term:
    """A symbolic term which consists of a variable with a given coefficient.

    Args:
        variable: The variable of the term.
        coefficient: The coefficient of the term.
    """

    def __eq__(self, arg0: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __init__(self, variable: Variable, coefficient: float = 1.0) -> None: ...
    def __mul__(self, arg0: float) -> Term: ...
    def __ne__(self, arg0: object) -> bool: ...
    def __rmul__(self, arg0: float) -> Term: ...
    def __rtruediv__(self, arg0: float) -> Term: ...
    def __truediv__(self, arg0: float) -> Term: ...
    def add_coefficient(self, coeff: float) -> None:
        """Add a coefficient to the coefficient of this term.

        Args:
            coeff: The coefficient to add.
        """

    def evaluate(self, assignment: Mapping[Variable, float]) -> float:
        """Evaluate the term with a given variable assignment.

        Args:
            assignment: The variable assignment.

        Returns:
            The evaluated value of the term.
        """

    def has_zero_coefficient(self) -> bool:
        """Check if the coefficient of the term is zero."""

    @property
    def coefficient(self) -> float:
        """The coefficient of the term."""

    @property
    def variable(self) -> Variable:
        """The variable of the term."""

class Expression:
    r"""A symbolic expression which consists of a sum of terms and a constant.

    The expression is of the form :math:`constant + term_1 + term_2 + \dots + term_n`.

    Args:
        terms: The list of terms.
        constant: The constant.

    Alternatively, an expression can be created with a single term and a constant or just a constant.
    """

    constant: float
    """
    The constant of the expression.
    """

    @overload
    def __add__(self, arg0: Expression) -> Expression: ...
    @overload
    def __add__(self, arg0: Term) -> Expression: ...
    @overload
    def __add__(self, arg0: float) -> Expression: ...
    def __eq__(self, arg0: object) -> bool: ...
    def __getitem__(self, idx: int) -> Term: ...
    def __hash__(self) -> int: ...
    @overload
    def __init__(self, terms: Sequence[Term], constant: float = 0.0) -> None:
        """Create an expression with a given list of terms and a constant.

        Args:
            terms: The list of terms.
            constant: The constant.
        """

    @overload
    def __init__(self, term: Term, constant: float = 0.0) -> None:
        """Create an expression with a given term and a constant.

        Args:
            term: The term.
            constant: The constant.
        """

    @overload
    def __init__(self, constant: float = 0.0) -> None:
        """Create an expression with a given constant.

        Args:
            constant: The constant.
        """

    def __iter__(self) -> Iterator[Term]: ...
    def __len__(self) -> int: ...
    def __mul__(self, arg0: float) -> Expression: ...
    def __ne__(self, arg0: object) -> bool: ...
    @overload
    def __radd__(self, arg0: Term) -> Expression: ...
    @overload
    def __radd__(self, arg0: float) -> Expression: ...
    def __rmul__(self, arg0: float) -> Expression: ...
    @overload
    def __rsub__(self, arg0: Term) -> Expression: ...
    @overload
    def __rsub__(self, arg0: float) -> Expression: ...
    def __rtruediv__(self, arg0: float) -> Expression: ...
    @overload
    def __sub__(self, arg0: Expression) -> Expression: ...
    @overload
    def __sub__(self, arg0: Term) -> Expression: ...
    @overload
    def __sub__(self, arg0: float) -> Expression: ...
    def __truediv__(self, arg0: float) -> Expression: ...
    def evaluate(self, assignment: Mapping[Variable, float]) -> float:
        """Evaluate the expression with a given variable assignment.

        Args:
            assignment: The variable assignment.

        Returns:
            The evaluated value of the expression.
        """

    def is_constant(self) -> bool:
        """Check if the expression is a constant."""

    def is_zero(self) -> bool:
        """Check if the expression is zero."""

    def num_terms(self) -> int:
        """The number of terms in the expression."""

    @property
    def terms(self) -> list[Term]:
        """The terms of the expression."""

    @property
    def variables(self) -> set[Variable]:
        """The variables in the expression."""
